# RUN: python %s | filecheck %s

from ctypes import c_float, c_int32, c_int64

from xdsl.dialects import arith, builtin
from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.program import FrontendProgram
from xdsl.frontend.pyast.utils.exceptions import CodeGenerationException


def add_i32(operand1: c_int32, operand2: c_int32) -> c_int32: ...
def and_i64(operand1: c_int64, operand2: c_int64) -> c_int64: ...
def mul_i32(operand1: c_int32, operand2: c_int32) -> c_int32: ...
def sub_i64(operand1: c_int64, operand2: c_int64) -> c_int64: ...
def shl_i32(operand1: c_int32, operand2: c_int32) -> c_int32: ...
def shrs_i64(operand1: c_int64, operand2: c_int64) -> c_int64: ...
def cmp_eq_i32(operand1: c_int32, operand2: c_int32) -> bool: ...
def cmp_le_i64(operand1: c_int64, operand2: c_int64) -> bool: ...
def cmp_lt_i32(operand1: c_int32, operand2: c_int32) -> bool: ...
def cmp_ge_i64(operand1: c_int64, operand2: c_int64) -> bool: ...
def cmp_gt_i32(operand1: c_int32, operand2: c_int32) -> bool: ...
def cmp_ne_i64(operand1: c_int64, operand2: c_int64) -> bool: ...
def add_f32(operand1: c_float, operand2: c_float) -> c_float: ...
def sub_f32(operand1: c_float, operand2: c_float) -> c_float: ...
def mul_f32(operand1: c_float, operand2: c_float) -> c_float: ...


p = FrontendProgram()
p.register_type(c_int32, builtin.i32)
p.register_type(c_int64, builtin.i64)
p.register_type(c_float, builtin.f32)
p.register_type(float, builtin.f64)
p.register_type(bool, builtin.i1)
p.register_function(add_i32, arith.AddiOp)
p.register_function(and_i64, arith.AndIOp)
p.register_function(mul_i32, arith.MuliOp)
p.register_function(sub_i64, arith.SubiOp)
p.register_function(shl_i32, arith.ShLIOp)
p.register_function(shrs_i64, arith.ShRSIOp)
p.register_function(cmp_eq_i32, lambda x, y: arith.CmpiOp(x, y, "eq"))  # pyright: ignore[reportUnknownArgumentType, reportUnknownLambdaType]
p.register_function(cmp_le_i64, lambda x, y: arith.CmpiOp(x, y, "sle"))  # pyright: ignore[reportUnknownArgumentType, reportUnknownLambdaType]
p.register_function(cmp_lt_i32, lambda x, y: arith.CmpiOp(x, y, "slt"))  # pyright: ignore[reportUnknownArgumentType, reportUnknownLambdaType]
p.register_function(cmp_ge_i64, lambda x, y: arith.CmpiOp(x, y, "sge"))  # pyright: ignore[reportUnknownArgumentType, reportUnknownLambdaType]
p.register_function(cmp_gt_i32, lambda x, y: arith.CmpiOp(x, y, "sgt"))  # pyright: ignore[reportUnknownArgumentType, reportUnknownLambdaType]
p.register_function(cmp_ne_i64, lambda x, y: arith.CmpiOp(x, y, "ne"))  # pyright: ignore[reportUnknownArgumentType, reportUnknownLambdaType]
p.register_function(add_f32, arith.AddfOp)
p.register_function(float.__add__, arith.AddfOp)
p.register_function(sub_f32, arith.SubfOp)
p.register_function(float.__sub__, arith.SubfOp)
p.register_function(mul_f32, arith.MulfOp)
p.register_function(float.__mul__, arith.MulfOp)
with CodeContext(p):
    # CHECK: arith.addi %{{.*}}, %{{.*}} : i32
    def test_addi_overload(a: c_int32, b: c_int32) -> c_int32:
        return add_i32(a, b)

    # CHECK: arith.andi %{{.*}}, %{{.*}} : i64
    def test_andi_overload(a: c_int64, b: c_int64) -> c_int64:
        return and_i64(a, b)

    # CHECK: arith.muli %{{.*}}, %{{.*}} : i32
    def test_muli_overload(a: c_int32, b: c_int32) -> c_int32:
        return mul_i32(a, b)

    # CHECK: arith.subi %{{.*}}, %{{.*}} : i64
    def test_subi_overload(a: c_int64, b: c_int64) -> c_int64:
        return sub_i64(a, b)

    # CHECK: arith.shli %{{.*}}, %{{.*}} : i32
    def test_shli_overload(a: c_int32, b: c_int32) -> c_int32:
        return shl_i32(a, b)

    # CHECK: arith.shrsi %{{.*}}, %{{.*}} : i64
    def test_shrsi_overload(a: c_int64, b: c_int64) -> c_int64:
        return shrs_i64(a, b)

    # CHECK:  arith.cmpi eq, %{{.*}}, %{{.*}} : i32
    def test_cmpi_eq_overload(a: c_int32, b: c_int32) -> bool:
        return cmp_eq_i32(a, b)

    # CHECK: arith.cmpi sle, %{{.*}}, %{{.*}} : i64
    def test_cmpi_le_overload(a: c_int64, b: c_int64) -> bool:
        return cmp_le_i64(a, b)

    # CHECK: arith.cmpi slt, %{{.*}}, %{{.*}} : i32
    def test_cmpi_lt_overload(a: c_int32, b: c_int32) -> bool:
        return cmp_lt_i32(a, b)

    # CHECK: arith.cmpi sge, %{{.*}}, %{{.*}} : i64
    def test_cmpi_ge_overload(a: c_int64, b: c_int64) -> bool:
        return cmp_ge_i64(a, b)

    # CHECK: arith.cmpi sgt, %{{.*}}, %{{.*}} : i32
    def test_cmpi_gt_overload(a: c_int32, b: c_int32) -> bool:
        return cmp_gt_i32(a, b)

    # CHECK: arith.cmpi ne, %{{.*}}, %{{.*}} : i64
    def test_cmpi_ne_overload(a: c_int64, b: c_int64) -> bool:
        return cmp_ne_i64(a, b)

    # CHECK: arith.addf %{{.*}}, %{{.*}} : f32
    def test_addf_overload_f32(a: c_float, b: c_float) -> c_float:
        return add_f32(a, b)

    # CHECK: arith.addf %{{.*}}, %{{.*}} : f64
    def test_addf_overload_f64(a: float, b: float) -> float:
        return a + b

    # CHECK: arith.subf %{{.*}}, %{{.*}} : f32
    def test_subf_overload_f32(a: c_float, b: c_float) -> c_float:
        return sub_f32(a, b)

    # CHECK: arith.subf %{{.*}}, %{{.*}} : f64
    def test_subf_overload_f64(a: float, b: float) -> float:
        return a - b

    # CHECK: arith.mulf %{{.*}}, %{{.*}} : f32
    def test_mulf_overload_f32(a: c_float, b: c_float) -> c_float:
        return mul_f32(a, b)

    # CHECK: arith.mulf %{{.*}}, %{{.*}} : f64
    def test_mulf_overload_f64(a: float, b: float) -> float:
        return a * b


p.compile(desymref=False)
print(p.textual_format())

try:
    with CodeContext(p):
        # CHECK: Binary operation 'FloorDiv' is not supported by type 'float' which does not overload '__floordiv__'.
        def test_missing_floordiv_overload_f64(a: float, b: float) -> float:
            # We expect the type error here, since the function doesn't exist on f64
            return a // b

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Comparison operation 'In' is not supported by type 'f64' which does not overload '__contains__'.
        def test_missing_contains_overload_f64(a: float, b: float) -> float:
            # We expect the type error here, since the function doesn't exist on f64
            return a in b  # pyright: ignore[reportOperatorIssue]

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)



from collections.abc import Callable, Sequence

from xdsl.backend.register_type import RegisterType
from xdsl.dialects.builtin import (
    UnrealizedConversionCastOp,
)
from xdsl.ir import Attribute, Operation, SSAValue


def cast_to_regs(
    values: Sequence[SSAValue],
    register_map: Callable[[Attribute], type[RegisterType]],
) -> tuple[list[Operation], list[SSAValue[Attribute]]]:
    """
    Return cast operations for operands that don't already have a register type
    and the new list of values that are all guaranteed to have register types.
    """
    registers: list[SSAValue] = []
    operations: list[Operation] = []
    for v in values:
        if isinstance(v.type, RegisterType):
            new_value = v
        else:
            cast_op, new_value = UnrealizedConversionCastOp.cast_one(
                v, register_map(v.type).unallocated()
            )
            new_value.name_hint = v.name_hint
            operations.append(cast_op)
        registers.append(new_value)
    return operations, registers



from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import affine, arith
from xdsl.dialects.builtin import IndexType, IntegerAttr, ModuleOp
from xdsl.dialects.scf import ParallelOp, ReduceOp
from xdsl.ir import Block, Operation, Region, SSAValue
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.hints import isa


@dataclass
class ScfParallelLoopTilingPattern(RewritePattern):
    tile_sizes: tuple[int, ...]

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ParallelOp, rewriter: PatternRewriter, /):
        # Only tile the innermost parallel loop
        if any(isinstance(o, ParallelOp) for o in op.body.walk()):
            return
        lower = op.lowerBound
        upper = op.upperBound
        step = op.step
        # The pass is meant to work on any parallel loop with any nu;ber of tile sizes.
        # For a loop of dimension N, either use the N first tile sizes or use them all
        # and fill the rest with 1.
        tile_sizes_v = self.tile_sizes[: len(lower)] + (1,) * (
            len(lower) - len(self.tile_sizes)
        )

        zero = arith.ConstantOp(IntegerAttr.from_index_int_value(0))
        tile_sizes_v = {i: s for i, s in enumerate(tile_sizes_v) if s != 0}
        tile_sizes = {
            i: arith.ConstantOp(IntegerAttr.from_index_int_value(s))
            for i, s in tile_sizes_v.items()
        }
        tiled_dims = sorted(tile_sizes.keys())
        if not tiled_dims:
            return
        outter_lower = [lower[d] for d in tiled_dims]
        outter_upper = [upper[d] for d in tiled_dims]
        outter_step = [arith.MuliOp(step[d], tile_sizes[d]) for d in tiled_dims]

        outter_loop = ParallelOp(
            outter_lower,
            outter_upper,
            outter_step,
            Region(
                Block(
                    [(outter_reduce := ReduceOp())],
                    arg_types=[IndexType()] * len(outter_lower),
                )
            ),
        )

        inner_lower = list[SSAValue | Operation]()
        inner_upper = list[SSAValue | Operation]()
        minops = list[Operation]()
        minmap = affine.AffineMapAttr(
            affine.AffineMap(
                3,
                0,
                (
                    affine.AffineExpr.dimension(0),
                    affine.AffineExpr.dimension(1) - affine.AffineExpr.dimension(2),
                ),
            )
        )
        for i in range(len(op.lowerBound)):
            if i in tile_sizes:
                inner_lower.append(zero)
                ilower, iupper, istep = lower[i], upper[i], step[i]
                if (
                    isinstance(ilower, arith.ConstantOp)
                    and isinstance(iupper, arith.ConstantOp)
                    and isinstance(istep, arith.ConstantOp)
                ):
                    lower_v, upper_v, step_v = (
                        c.value for c in (ilower, iupper, istep)
                    )
                    assert isa(lower_v, IntegerAttr[IndexType])
                    assert isa(upper_v, IntegerAttr[IndexType])
                    assert isa(step_v, IntegerAttr[IndexType])
                    lower_v, upper_v, step_v = (
                        c.value.data for c in (lower_v, upper_v, step_v)
                    )
                    iter_count = (upper_v - lower_v) // step_v
                    if iter_count % tile_sizes_v[i] == 0:
                        inner_upper.append(tile_sizes[i])
                        continue

                arg_index = tiled_dims.index(i)
                minop = affine.MinOp(
                    operands=[
                        [
                            tile_sizes[i],
                            outter_upper[arg_index],
                            outter_loop.body.block.args[arg_index],
                        ]
                    ],
                    properties={"map": minmap},
                    result_types=[IndexType()],
                )
                minops.append(minop)
                inner_upper.append(minop)

            else:
                inner_lower.append(lower[i])
                inner_upper.append(upper[i])

        inner_loop = ParallelOp(inner_lower, inner_upper, step, op.detach_region(0))
        for i, arg in reversed(list(enumerate(inner_loop.body.block.args))):
            if i in tile_sizes:
                arg_index = tiled_dims.index(i)
                iv = arith.AddiOp(outter_loop.body.block.args[arg_index], arg)
                assert inner_loop.body.block.first_op is not None
                inner_loop.body.block.insert_op_before(
                    iv, inner_loop.body.block.first_op
                )
                for use in tuple(arg.uses):
                    if use.operation is iv:
                        continue
                    use.operation.operands[use.index] = iv.result
        outter_loop.body.block.insert_ops_before([*minops, inner_loop], outter_reduce)
        rewriter.replace_matched_op(
            [zero, *tile_sizes.values(), *outter_step, outter_loop]
        )


@dataclass(frozen=True)
class ScfParallelLoopTilingPass(ModulePass):
    name = "scf-parallel-loop-tiling"

    parallel_loop_tile_sizes: tuple[int, ...]

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        walker = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [ScfParallelLoopTilingPattern(tuple(self.parallel_loop_tile_sizes))]
            ),
            walk_regions_first=True,
            apply_recursively=False,
        )
        walker.rewrite_module(op)



import pytest

from xdsl.builder import ImplicitBuilder
from xdsl.dialects.builtin import i1
from xdsl.dialects.test import TestOp
from xdsl.ir import Block, Region
from xdsl.ir.region_utils import used_values_defined_above


def test_used_values_defined_above():
    r = Region(Block())
    r0 = Region(Block())
    r1 = Region(Block())
    r00 = Region(Block())
    r01 = Region(Block())

    with ImplicitBuilder(r.block):
        (v, _) = TestOp(result_types=(i1, i1)).results
        TestOp(regions=(r0, r1))
        with ImplicitBuilder(r0):
            (v0, _) = TestOp(operands=(v,), result_types=(i1, i1)).results
            TestOp(regions=(r00, r01))
            with ImplicitBuilder(r00):
                (v00, _) = TestOp(
                    operands=(
                        v,
                        v0,
                    ),
                    result_types=(i1, i1),
                ).results
                TestOp(operands=(v00,))
            with ImplicitBuilder(r01):
                (v01, _) = TestOp(
                    operands=(
                        v,
                        v0,
                    ),
                    result_types=(i1, i1),
                ).results
                TestOp(operands=(v01,))
        with ImplicitBuilder(r1):
            (v1, _) = TestOp(operands=(v,), result_types=(i1, i1)).results
            TestOp(operands=(v1,))

    assert used_values_defined_above(r) == set()
    assert used_values_defined_above(r0) == {v}
    assert used_values_defined_above(r1) == {v}
    assert used_values_defined_above(r00) == {v, v0}
    assert used_values_defined_above(r01) == {v, v0}

    assert used_values_defined_above(r00, r0) == {v}
    assert used_values_defined_above(r01, r0) == {v}

    with pytest.raises(
        AssertionError,
        match="expected isolation limit to be an ancestor of the given region",
    ):
        used_values_defined_above(r0, r1)



"""
Helpers for encoding modules in the `wasm` dialect to the WebAssembly binary format.
"""

import abc
from typing import BinaryIO


class WasmBinaryEncodingContext:
    """
    A class to store the state of encoding.
    """


class EncodingException(Exception): ...


class WasmBinaryEncodable(abc.ABC):
    @abc.abstractmethod
    def encode(self, ctx: WasmBinaryEncodingContext, io: BinaryIO) -> None:
        raise NotImplementedError()



from xdsl.dialects.x86 import GeneralRegisterType
from xdsl.dialects.x86.ops import (
    DI_Operation,
    DM_Operation,
    DMI_Operation,
    DS_Operation,
    DSI_Operation,
    M_Operation,
    MI_Operation,
    MS_Operation,
    R_Operation,
    RI_Operation,
    RM_Operation,
    RS_Operation,
    RSS_Operation,
)
from xdsl.irdl import irdl_op_definition
from xdsl.utils import test_value

reg_type = GeneralRegisterType.infinite_register


@irdl_op_definition
class TestRSOperation(RS_Operation[GeneralRegisterType, GeneralRegisterType]):
    """Test operation that inherits from RS_Operation for testing register constraints."""

    name = "test.rs_operation"


def test_rs_operation_register_constraints():
    rs_op = TestRSOperation(
        test_value.create_ssa_value(reg_type(0)),
        test_value.create_ssa_value(reg_type(1)),
    )

    rs_c = rs_op.get_register_constraints()

    assert rs_c.ins == (rs_op.source,)
    assert rs_c.outs == ()
    assert rs_c.inouts == ((rs_op.register_in, rs_op.register_out),)


@irdl_op_definition
class TestROperation(R_Operation[GeneralRegisterType]):
    """Test operation that inherits from R_Operation for testing register constraints."""

    name = "test.r_operation"


def test_r_operation_register_constraints():
    r_op = TestROperation(
        test_value.create_ssa_value(reg_type(0)),
    )

    r_c = r_op.get_register_constraints()

    assert r_c.ins == ()
    assert r_c.outs == ()
    assert r_c.inouts == ((r_op.register_in, r_op.register_out),)


@irdl_op_definition
class TestRMOperation(RM_Operation[GeneralRegisterType, GeneralRegisterType]):
    """Test operation that inherits from RM_Operation for testing register constraints."""

    name = "test.rm_operation"


def test_rm_operation_register_constraints():
    # Create an instance of our test RM_Operation
    rm_op = TestRMOperation(
        test_value.create_ssa_value(reg_type(0)),
        test_value.create_ssa_value(reg_type(1)),
        memory_offset=42,
        register_out=reg_type(2),
    )

    rm_c = rm_op.get_register_constraints()

    assert rm_c.ins == (rm_op.memory,)
    assert rm_c.outs == ()
    assert rm_c.inouts == ((rm_op.register_in, rm_op.register_out),)


@irdl_op_definition
class TestDMOperation(DM_Operation[GeneralRegisterType, GeneralRegisterType]):
    """Test operation that inherits from DM_Operation for testing register constraints."""

    name = "test.dm_operation"


def test_dm_operation_register_constraints():
    # Create an instance of our test DM_Operation
    dm_op = TestDMOperation(
        test_value.create_ssa_value(reg_type(0)),
        memory_offset=42,
        destination=reg_type(1),
    )

    dm_c = dm_op.get_register_constraints()

    assert tuple(dm_c.ins) == (dm_op.memory,)
    assert dm_c.outs == (dm_op.destination,)
    assert dm_c.inouts == ()


@irdl_op_definition
class TestRIOperation(RI_Operation[GeneralRegisterType]):
    """Test operation that inherits from RI_Operation for testing register constraints."""

    name = "test.ri_operation"


def test_ri_operation_register_constraints():
    # Create an instance of our test RI_Operation
    ri_op = TestRIOperation(
        test_value.create_ssa_value(reg_type(0)),
        immediate=42,
        register_out=reg_type(1),
    )

    ri_c = ri_op.get_register_constraints()

    assert ri_c.ins == ()
    assert ri_c.outs == ()
    assert ri_c.inouts == ((ri_op.register_in, ri_op.register_out),)


@irdl_op_definition
class TestMSOperation(MS_Operation[GeneralRegisterType, GeneralRegisterType]):
    """Test operation that inherits from MS_Operation for testing register constraints."""

    name = "test.ms_operation"


def test_ms_operation_register_constraints():
    # Create an instance of our test MS_Operation
    ms_op = TestMSOperation(
        test_value.create_ssa_value(reg_type(0)),
        test_value.create_ssa_value(reg_type(1)),
        memory_offset=42,
    )

    ms_c = ms_op.get_register_constraints()

    assert tuple(ms_c.ins) == (
        ms_op.memory,
        ms_op.source,
    )
    assert ms_c.outs == ()
    assert ms_c.inouts == ()


@irdl_op_definition
class TestMIOperation(MI_Operation[GeneralRegisterType]):
    """Test operation that inherits from MI_Operation for testing register constraints."""

    name = "test.mi_operation"


def test_mi_operation_register_constraints():
    # Create an instance of our test MI_Operation
    mi_op = TestMIOperation(
        test_value.create_ssa_value(reg_type(0)),
        immediate=42,
        memory_offset=10,
    )

    mi_c = mi_op.get_register_constraints()

    assert tuple(mi_c.ins) == (mi_op.memory,)
    assert mi_c.outs == ()
    assert mi_c.inouts == ()


@irdl_op_definition
class TestDSIOperation(DSI_Operation[GeneralRegisterType, GeneralRegisterType]):
    """Test operation that inherits from DSI_Operation for testing register constraints."""

    name = "test.dsi_operation"


def test_dsi_operation_register_constraints():
    # Create an instance of our test DSI_Operation
    dsi_op = TestDSIOperation(
        test_value.create_ssa_value(reg_type(0)),
        immediate=42,
        destination=reg_type(1),
    )

    dsi_c = dsi_op.get_register_constraints()

    assert tuple(dsi_c.ins) == (dsi_op.source,)
    assert dsi_c.outs == (dsi_op.destination,)
    assert dsi_c.inouts == ()


@irdl_op_definition
class TestDMIOperation(DMI_Operation[GeneralRegisterType, GeneralRegisterType]):
    """Test operation that inherits from DMI_Operation for testing register constraints."""

    name = "test.dmi_operation"


def test_dmi_operation_register_constraints():
    # Create an instance of our test DMI_Operation
    dmi_op = TestDMIOperation(
        test_value.create_ssa_value(reg_type(0)),
        memory_offset=10,
        immediate=42,
        destination=reg_type(1),
    )

    dmi_c = dmi_op.get_register_constraints()

    assert tuple(dmi_c.ins) == (dmi_op.memory,)
    assert dmi_c.outs == (dmi_op.destination,)
    assert dmi_c.inouts == ()


@irdl_op_definition
class TestMOperation(M_Operation[GeneralRegisterType]):
    """Test operation that inherits from M_Operation for testing register constraints."""

    name = "test.m_operation"


def test_m_operation_register_constraints():
    # Create an instance of our test M_Operation
    m_op = TestMOperation(
        test_value.create_ssa_value(reg_type(0)),
        memory_offset=10,
    )

    m_c = m_op.get_register_constraints()

    assert tuple(m_c.ins) == (m_op.memory,)
    assert m_c.outs == ()
    assert m_c.inouts == ()


@irdl_op_definition
class TestRSSOperation(
    RSS_Operation[GeneralRegisterType, GeneralRegisterType, GeneralRegisterType]
):
    """Test operation that inherits from RSS_Operation for testing register constraints."""

    name = "test.rss_operation"


def test_rss_operation_register_constraints():
    # Create an instance of our test RSS_Operation
    rss_op = TestRSSOperation(
        test_value.create_ssa_value(reg_type(0)),
        test_value.create_ssa_value(reg_type(1)),
        test_value.create_ssa_value(reg_type(2)),
    )

    rss_c = rss_op.get_register_constraints()

    assert tuple(rss_c.ins) == (rss_op.source1, rss_op.source2)
    assert rss_c.outs == ()
    assert rss_c.inouts == ((rss_op.register_in, rss_op.register_out),)


@irdl_op_definition
class TestDSOperation(DS_Operation[GeneralRegisterType, GeneralRegisterType]):
    """Test operation that inherits from DS_Operation for testing register constraints."""

    name = "test.ds_operation"


def test_ds_operation_register_constraints():
    # Create an instance of our test DS_Operation
    ds_op = TestDSOperation(
        test_value.create_ssa_value(reg_type(0)),
        destination=reg_type(1),
    )

    ds_c = ds_op.get_register_constraints()

    assert tuple(ds_c.ins) == (ds_op.source,)
    assert ds_c.outs == (ds_op.destination,)
    assert ds_c.inouts == ()


@irdl_op_definition
class TestDIOperation(DI_Operation[GeneralRegisterType]):
    """Test operation that inherits from DI_Operation for testing register constraints."""

    name = "test.di_operation"


def test_di_operation_register_constraints():
    # Create an instance of our test DI_Operation
    di_op = TestDIOperation(
        10,
        destination=reg_type(0),
    )

    di_c = di_op.get_register_constraints()

    assert tuple(di_c.ins) == ()
    assert di_c.outs == (di_op.destination,)
    assert di_c.inouts == ()



from dataclasses import dataclass

from xdsl.backend.x86.lowering.helpers import cast_operands_to_regs
from xdsl.context import Context
from xdsl.dialects import arith, builtin, x86
from xdsl.dialects.builtin import (
    IntegerAttr,
    UnrealizedConversionCastOp,
)
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.exceptions import DiagnosticException
from xdsl.utils.hints import isa


@dataclass
class ArithConstantToX86(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.ConstantOp, rewriter: PatternRewriter):
        if not isa(op.value, IntegerAttr):
            raise DiagnosticException(
                "Lowering of arith.constant is only implemented for integers"
            )
        mov_op = x86.DI_MovOp(
            immediate=op.value.value.data, destination=x86.register.UNALLOCATED_GENERAL
        )
        cast_op, _ = UnrealizedConversionCastOp.cast_one(
            mov_op.destination, op.result.type
        )
        rewriter.replace_matched_op([mov_op, cast_op])


@dataclass
class ArithAddiToX86(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.AddiOp, rewriter: PatternRewriter):
        if isinstance(op.lhs.type, builtin.ShapedType):
            raise DiagnosticException(
                "Lowering of arith.addi not implemented for ShapedType"
            )
        lhs_x86, rhs_x86 = cast_operands_to_regs(rewriter=rewriter)
        rhs_copy_op = x86.DS_MovOp(
            source=rhs_x86, destination=x86.register.UNALLOCATED_GENERAL
        )
        add_op = x86.RS_AddOp(source=lhs_x86, register_in=rhs_copy_op.destination)
        result_cast_op, _ = UnrealizedConversionCastOp.cast_one(
            add_op.register_out, op.lhs.type
        )
        rewriter.replace_matched_op([rhs_copy_op, add_op, result_cast_op])


@dataclass
class ArithMuliToX86(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.MuliOp, rewriter: PatternRewriter):
        if isinstance(op.lhs.type, builtin.ShapedType):
            raise DiagnosticException(
                "Lowering of arith.muli not implemented for ShapedType"
            )
        lhs_x86, rhs_x86 = cast_operands_to_regs(rewriter=rewriter)
        rhs_copy_op = x86.DS_MovOp(
            source=rhs_x86, destination=x86.register.UNALLOCATED_GENERAL
        )
        add_op = x86.RS_ImulOp(source=lhs_x86, register_in=rhs_copy_op.destination)
        result_cast_op, _ = UnrealizedConversionCastOp.cast_one(
            add_op.register_out, op.lhs.type
        )
        rewriter.replace_matched_op([rhs_copy_op, add_op, result_cast_op])


@dataclass(frozen=True)
class ConvertArithToX86Pass(ModulePass):
    name = "convert-arith-to-x86"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ArithAddiToX86(),
                    ArithMuliToX86(),
                    ArithConstantToX86(),
                ]
            ),
            apply_recursively=False,
        ).rewrite_module(op)



import json
from collections import defaultdict
from typing import Any, cast

from xdsl.context import Context
from xdsl.dialects import builtin, eqsat
from xdsl.ir import BlockArgument, Operation
from xdsl.passes import ModulePass


class _IDGenerator:
    def __init__(self, prefix: str):
        self.prefix = prefix
        self.counter = 0

    def __call__(self):
        self.counter += 1
        return f"{self.prefix}{self.counter}"


def serialize_to_egraph(mod: builtin.ModuleOp):
    enode_to_id: defaultdict[Operation | BlockArgument, str] = defaultdict(
        _IDGenerator("enode_")
    )
    eclass_to_id: defaultdict[eqsat.EClassOp, str] = defaultdict(
        _IDGenerator("eclass_")
    )
    nodes: dict[str, dict[str, str | list[str]]] = dict()
    for op in mod.walk(reverse=True):
        if isinstance(op, eqsat.EClassOp):
            for operand in op.operands:
                if isinstance(operand, BlockArgument):
                    nodes[enode_to_id[operand]] = {
                        "op": f"arg {operand.index}",
                        "eclass": eclass_to_id[op],
                        "children": [],
                    }
            continue
        children: list[Any] = []
        eclass_id = None
        for res in op.results:
            for use in res.uses:
                if isinstance(use.operation, eqsat.EClassOp):
                    assert len(op.results) == 1, (
                        "Only single result operations are supported"
                    )
                    assert res.has_one_use(), "Only single use operations are supported"
                    eclass_id = eclass_to_id[use.operation]
        if eclass_id is None:
            continue
        for operand in op.operands:
            if isinstance(operand.owner, eqsat.EClassOp):
                firstoperand = operand.owner.operands[0]
                if isinstance(firstoperand.owner, Operation):
                    canonical_enode = firstoperand.owner
                else:
                    canonical_enode = cast(BlockArgument, firstoperand)
                children.append(enode_to_id[canonical_enode])
        if op.operands:
            name = op.name
        else:
            # If the operation has no operands, we get the full string representation as name for the node.
            # This is useful for operations such as `arith.constant 42`.
            name = str(op).split("=")[1].split(":")[0].strip()
        nodes[enode_to_id[op]] = {
            "op": name,
            "eclass": eclass_id,
            "children": children,
        }
    return nodes


class SerializeEGraph(ModulePass):
    name = "eqsat-serialize-egraph"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        print(json.dumps({"nodes": serialize_to_egraph(op)}))



from contextlib import redirect_stdout
from io import StringIO

import pytest

from xdsl.builder import Builder
from xdsl.context import Context
from xdsl.dialects import riscv, riscv_debug, riscv_func
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import BlockArgument
from xdsl.transforms.riscv_register_allocation import RISCVRegisterAllocation

pytest.importorskip("riscemu", reason="riscemu is an optional dependency")

from xdsl.interpreters.riscv_emulator import RV_Debug, run_riscv  # noqa: E402

ctx = Context()
ctx.load_dialect(riscv.RISCV)


def test_simple():
    @ModuleOp
    @Builder.implicit_region
    def module():
        @Builder.implicit_region
        def body():
            six = riscv.LiOp(6).rd
            seven = riscv.LiOp(7).rd
            forty_two = riscv.MulOp(six, seven).rd
            riscv_debug.PrintfOp("{}", (forty_two,))
            riscv.ReturnOp()

        riscv_func.FuncOp("main", body, ((), ()), visibility="public")

    RISCVRegisterAllocation().apply(ctx, module)

    code = riscv.riscv_code(module)

    with StringIO() as stream, redirect_stdout(stream):
        run_riscv(
            code,
            extensions=[RV_Debug],
            unlimited_regs=True,
            verbosity=0,
        )
        assert stream.getvalue() == "42\n"


def test_multiply_add():
    @ModuleOp
    @Builder.implicit_region
    def module():
        @Builder.implicit_region
        def main():
            riscv.LiOp(3, rd=riscv.Registers.A0)
            riscv.LiOp(2, rd=riscv.Registers.A1)
            riscv.LiOp(1, rd=riscv.Registers.A2)

            riscv.JalOp("muladd")
            res = riscv.GetRegisterOp(riscv.Registers.A0).res
            riscv_debug.PrintfOp("{}", (res,))

            riscv.LiOp(93, rd=riscv.Registers.A7)
            riscv.EcallOp()

        riscv_func.FuncOp("main", main, ((), ()), visibility="public")

        @Builder.implicit_region((riscv.Registers.A0, riscv.Registers.A1))
        def multiply(args: tuple[BlockArgument, ...]):
            riscv.CommentOp("no extra registers needed, so no need to deal with stack")
            rs1, rs2 = args
            res = riscv.MulOp(rs1, rs2, rd=riscv.Registers.A0).rd
            riscv_func.ReturnOp(res)

        riscv_func.FuncOp(
            "multiply",
            multiply,
            ((riscv.Registers.A0, riscv.Registers.A1), (riscv.Registers.A0,)),
        )

        @Builder.implicit_region((riscv.Registers.A0, riscv.Registers.A1))
        def add(args: tuple[BlockArgument, ...]):
            riscv.CommentOp("no extra registers needed, so no need to deal with stack")
            rs1, rs2 = args
            res = riscv.AddOp(rs1, rs2, rd=riscv.Registers.A0).rd
            riscv_func.ReturnOp(res)

        riscv_func.FuncOp(
            "add",
            add,
            ((riscv.Registers.A0, riscv.Registers.A1), (riscv.Registers.A0,)),
        )

        @Builder.implicit_region(
            (riscv.Registers.A0, riscv.Registers.A1, riscv.Registers.A2)
        )
        def muladd(args: tuple[BlockArgument, ...]):
            riscv.CommentOp("a0 <- a0 * a1 + a2")
            riscv.CommentOp("prologue")
            # get registers with the arguments to muladd
            _, _, a2_muladd = args

            # get registers we'll use in this section
            sp_muladd = riscv.GetRegisterOp(riscv.Registers.SP)
            s0_muladd_0 = riscv.GetRegisterOp(riscv.Registers.S0)
            ra_muladd = riscv.GetRegisterOp(riscv.Registers.RA)
            riscv.CommentOp(
                "decrement stack pointer by number of register values we need to store for later"
            )
            riscv.AddiOp(sp_muladd, -8, rd=riscv.Registers.SP)
            riscv.CommentOp("save the s registers we'll use on the stack")
            riscv.SwOp(sp_muladd, s0_muladd_0, 0)
            riscv.CommentOp("save the return address we'll use on the stack")
            riscv.SwOp(sp_muladd, ra_muladd, 4)

            # store third parameter, in a2 to the temporary register s0
            # guaranteed to be the same after call to multiply
            s0_muladd_1 = riscv.MVOp(a2_muladd, rd=riscv.Registers.S0)
            riscv.JalOp("multiply")

            # The product of a0 and a1 is stored in a0
            # We now have to move s0 to a1, and call add

            riscv.MVOp(s0_muladd_1, rd=riscv.Registers.A1)

            riscv.JalOp("add")

            riscv.CommentOp("epilogue")
            riscv.CommentOp("store the old values back into the s registers")
            riscv.LwOp(sp_muladd, 0, rd=riscv.Registers.S0)

            riscv.CommentOp("store the return address back into the ra register")
            riscv.LwOp(sp_muladd, 4, rd=riscv.Registers.RA)

            riscv.CommentOp(
                "set the sp back to what it was at the start of the function call"
            )
            riscv.AddiOp(sp_muladd, 8, rd=riscv.Registers.SP)

            riscv.CommentOp("jump back to caller")
            riscv.ReturnOp()

        riscv_func.FuncOp(
            "muladd",
            muladd,
            (
                (riscv.Registers.A0, riscv.Registers.A1, riscv.Registers.A2),
                (riscv.Registers.A0,),
            ),
        )

    RISCVRegisterAllocation().apply(ctx, module)

    code = riscv.riscv_code(module)
    with StringIO() as stream, redirect_stdout(stream):
        run_riscv(
            code,
            extensions=[RV_Debug],
            unlimited_regs=True,
            verbosity=0,
        )
        assert stream.getvalue() == "7\n"



"""
mod_arith is a dialect implementing modular arithmetic, originally
implemented as part of the HEIR project (https://github.com/google/heir/tree/main).
"""

from abc import ABC
from typing import ClassVar

from xdsl.dialects.arith import signlessIntegerLike
from xdsl.dialects.builtin import IntegerAttr
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.irdl import (
    IRDLOperation,
    ParsePropInAttrDict,
    VarConstraint,
    irdl_op_definition,
    operand_def,
    prop_def,
    result_def,
    traits_def,
)
from xdsl.traits import Pure


class BinaryOp(IRDLOperation, ABC):
    """
    Simple binary operation
    """

    T: ClassVar = VarConstraint("T", signlessIntegerLike)

    lhs = operand_def(T)
    rhs = operand_def(T)
    output = result_def(T)
    modulus = prop_def(IntegerAttr)

    irdl_options = [ParsePropInAttrDict()]

    assembly_format = "$lhs `,` $rhs attr-dict `:` type($output)"
    traits = traits_def(Pure())

    def __init__(
        self,
        lhs: SSAValue | Operation,
        rhs: SSAValue | Operation,
        result_type: Attribute | None,
        modulus: Attribute,
    ):
        if result_type is None:
            result_type = SSAValue.get(lhs).type

        super().__init__(
            operands=[lhs, rhs],
            result_types=[result_type],
            properties={"modulus": modulus},
        )


@irdl_op_definition
class AddOp(BinaryOp):
    name = "mod_arith.add"


ModArith = Dialect(
    "mod_arith",
    [
        AddOp,
    ],
)



from xdsl.dialects.builtin import (
    TensorType,
    f32,
    i32,
)
from xdsl.dialects.tosa import are_tosa_broadcastable
from xdsl.utils.test_value import create_ssa_value

t_i = TensorType(i32, [1, 2, 3, 4])
t_f = TensorType(f32, [1, 2, 3, 4])
t_flat = TensorType(i32, [1, 1, 1, 1])
t_small = TensorType(i32, [1, 2])
t_large = TensorType(i32, [4, 5, 6, 7])

tensor_i = create_ssa_value(t_i)
tensor_f = create_ssa_value(t_f)
tensor_flat = create_ssa_value(t_flat)
tensor_small = create_ssa_value(t_small)
tensor_large = create_ssa_value(t_large)


def test_are_tosa_broadcastable():
    # test same
    assert are_tosa_broadcastable(t_i, t_i, t_i)

    # test broadcasting
    assert are_tosa_broadcastable(t_i, t_flat, t_i)
    assert are_tosa_broadcastable(t_flat, t_i, t_i)
    assert not are_tosa_broadcastable(t_i, t_flat, t_flat)

    # test shape mismatch
    assert not are_tosa_broadcastable(t_i, t_small, t_i)

    # test mismatched dim sizes
    assert not are_tosa_broadcastable(t_i, t_i, t_large)



import pytest

from xdsl.dialects.arith import (
    AddfOp,
    AddiOp,
    AddUIExtendedOp,
    AndIOp,
    BitcastOp,
    CeilDivSIOp,
    CeilDivUIOp,
    CmpfOp,
    CmpiOp,
    ConstantOp,
    DivfOp,
    DivSIOp,
    DivUIOp,
    ExtFOp,
    ExtSIOp,
    ExtUIOp,
    FastMathFlagsAttr,
    FloatingPointLikeBinaryOperation,
    FloorDivSIOp,
    FPToSIOp,
    FPToUIOp,
    IndexCastOp,
    MaximumfOp,
    MaxSIOp,
    MaxUIOp,
    MinimumfOp,
    MinSIOp,
    MinUIOp,
    MulfOp,
    MulSIExtendedOp,
    MulUIExtendedOp,
    NegfOp,
    OrIOp,
    RemSIOp,
    RemUIOp,
    SelectOp,
    ShLIOp,
    ShRSIOp,
    ShRUIOp,
    SignlessIntegerBinaryOperation,
    SIToFPOp,
    SubfOp,
    SubiOp,
    TruncFOp,
    TruncIOp,
    UIToFPOp,
    XOrIOp,
)
from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
    FloatAttr,
    IndexType,
    IntegerAttr,
    IntegerType,
    MemRefType,
    Signedness,
    TensorType,
    VectorType,
    f32,
    f64,
    i1,
    i32,
    i64,
)
from xdsl.ir import Attribute
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.test_value import create_ssa_value


class Test_integer_arith_construction:
    operand_type = i32
    a = ConstantOp.from_int_and_width(1, operand_type)
    b = ConstantOp.from_int_and_width(1, operand_type)

    @pytest.mark.parametrize(
        "OpClass",
        [
            AddiOp,
            SubiOp,
            DivUIOp,
            DivSIOp,
            FloorDivSIOp,
            CeilDivSIOp,
            CeilDivUIOp,
            RemUIOp,
            RemSIOp,
            MinUIOp,
            MinSIOp,
            MaxUIOp,
            MaxSIOp,
            AndIOp,
            OrIOp,
            XOrIOp,
            ShLIOp,
            ShRUIOp,
            ShRSIOp,
        ],
    )
    @pytest.mark.parametrize("return_type", [None, operand_type])
    def test_arith_ops_init(
        self,
        OpClass: type[SignlessIntegerBinaryOperation],
        return_type: Attribute,
    ):
        op = OpClass(self.a, self.b)

        assert isinstance(op, OpClass)
        assert op.lhs.owner is self.a
        assert op.rhs.owner is self.b
        assert op.result.type == self.operand_type

    def test_Cmpi(self):
        _ = CmpiOp(self.a, self.b, 2)

    @pytest.mark.parametrize(
        "input",
        ["eq", "ne", "slt", "sle", "ult", "ule", "ugt", "uge"],
    )
    def test_Cmpi_from_mnemonic(self, input: str):
        _ = CmpiOp(self.a, self.b, input)


def test_constant_construction():
    c1 = ConstantOp(IntegerAttr(1, i32))
    assert c1.value.type == i32

    c3 = ConstantOp(FloatAttr(1.0, f32))
    assert c3.value.type == f32

    value_type = TensorType(i32, [2, 2])
    c5 = ConstantOp(DenseIntOrFPElementsAttr.from_list(value_type, [1, 2, 3, 4]))
    assert c5.value.type == value_type


@pytest.mark.parametrize(
    "value, truncated",
    [
        (-1, -1),
        (1, 1),
        (255, -1),
        (256, 0),
    ],
)
def test_constant_truncation(value: int, truncated: int):
    constant = ConstantOp.from_int_and_width(value, 8, truncate_bits=True)
    assert isinstance(v := constant.value, IntegerAttr)
    assert v.value.data == truncated


@pytest.mark.parametrize(
    "lhs_type, rhs_type, sum_type, is_correct",
    [
        (i32, i32, None, True),
        (i32, i32, i32, True),
        (i32, i32, i64, False),
        (i32, i64, None, False),
        (VectorType(i32, [4]), VectorType(i32, [4]), None, True),
        (VectorType(i32, [4]), VectorType(i32, [5]), None, False),
        (VectorType(i32, [4]), VectorType(i64, [5]), None, False),
        (VectorType(i32, [4]), VectorType(i32, [4]), VectorType(i32, [4]), True),
        (VectorType(i32, [4]), VectorType(i32, [4]), VectorType(i64, [4]), False),
        (TensorType(i32, [4]), TensorType(i32, [4]), None, True),
        (TensorType(i32, [4]), TensorType(i32, [5]), None, False),
        (TensorType(i32, [4]), TensorType(i64, [5]), None, False),
        (TensorType(i32, [4]), TensorType(i32, [4]), TensorType(i32, [4]), True),
        (TensorType(i32, [4]), TensorType(i32, [4]), TensorType(i1, [4]), False),
        (VectorType(i32, [4]), TensorType(i32, [4]), None, False),
        (VectorType(i32, [4]), TensorType(i32, [4]), TensorType(i32, [4]), False),
    ],
)
def test_addui_extend(
    lhs_type: Attribute,
    rhs_type: Attribute,
    sum_type: Attribute | None,
    is_correct: bool,
):
    lhs = create_ssa_value(lhs_type)
    rhs = create_ssa_value(rhs_type)

    attributes = {"foo": i32}

    if is_correct:
        op = AddUIExtendedOp(lhs, rhs, attributes, sum_type)
        op.verify()
        assert op.lhs == lhs
        assert op.rhs == rhs
        assert op.attributes == attributes
        if sum_type:
            assert op.sum.type == sum_type
        assert op.overflow.type == AddUIExtendedOp.infer_overflow_type(lhs_type)
        if isinstance(container_type := op.overflow.type, VectorType | TensorType):
            assert container_type.element_type == i1
        else:
            assert op.overflow.type == i1
    else:
        with pytest.raises((VerifyException, ValueError)):
            op = AddUIExtendedOp(lhs, rhs, attributes, sum_type)
            op.verify()


@pytest.mark.parametrize("op_type", [MulSIExtendedOp, MulUIExtendedOp])
def test_mul_extended(op_type: type[MulSIExtendedOp | MulUIExtendedOp]):
    lhs = create_ssa_value(i32)
    rhs = create_ssa_value(i32)

    op = op_type(lhs, rhs)

    assert op.lhs == lhs
    assert op.rhs == rhs
    assert op.low.type == i32
    assert op.high.type == i32

    op2 = op_type(lhs, rhs, i64)

    assert op2.lhs == lhs
    assert op2.rhs == rhs
    assert op2.low.type == i64
    assert op2.high.type == i64


class Test_float_arith_construction:
    a = ConstantOp(FloatAttr(1.1, f32))
    b = ConstantOp(FloatAttr(2.2, f32))

    @pytest.mark.parametrize(
        "func",
        [AddfOp, SubfOp, MulfOp, DivfOp, MaximumfOp, MinimumfOp],
    )
    @pytest.mark.parametrize(
        "flags", [FastMathFlagsAttr("none"), FastMathFlagsAttr("fast"), None]
    )
    def test_arith_ops(
        self,
        func: type[FloatingPointLikeBinaryOperation],
        flags: FastMathFlagsAttr | None,
    ):
        op = func(self.a, self.b, flags)
        assert op.operands[0].owner is self.a
        assert op.operands[1].owner is self.b
        assert op.fastmath == (flags or FastMathFlagsAttr("none"))


def test_select_op():
    t = ConstantOp.from_int_and_width(1, IntegerType(1))
    f = ConstantOp.from_int_and_width(0, IntegerType(1))
    select_t_op = SelectOp(t, t, f)
    select_f_op = SelectOp(f, t, f)
    select_t_op.verify_()
    select_f_op.verify_()

    # wanting to verify it actually selected the correct operand, but not sure if in correct scope
    assert select_t_op.result.type == t.result.type
    assert select_f_op.result.type == f.result.type


@pytest.mark.parametrize(
    "in_type, out_type",
    [
        (i1, IntegerType(1, signedness=Signedness.UNSIGNED)),
        (i32, f32),
        (i64, f64),
        (i32, i32),
        (IndexType(), i1),
        (i1, IndexType()),
        (f32, IndexType()),
        (IndexType(), f64),
        (VectorType(i64, [3]), VectorType(f64, [3])),
        (VectorType(f32, [3]), VectorType(i32, [3])),
        (MemRefType(i32, [5]), MemRefType(f32, [5])),
    ],
)
def test_bitcast_op(in_type: Attribute, out_type: Attribute):
    in_arg = create_ssa_value(in_type)
    cast = BitcastOp(in_arg, out_type)

    cast.verify_()
    assert cast.result.type == out_type


SHAPE_MISMATCH = "operand and result type must have compatible shape"
BITWIDTH_MISMATCH = "operand and result types must have equal bitwidths or be IndexType"


@pytest.mark.parametrize(
    "in_type, out_type, err_msg",
    [
        (i1, i32, BITWIDTH_MISMATCH),
        (i32, i64, BITWIDTH_MISMATCH),
        (i64, i32, BITWIDTH_MISMATCH),
        (f32, i64, BITWIDTH_MISMATCH),
        (f32, f64, BITWIDTH_MISMATCH),
        (VectorType(i32, [5]), i32, SHAPE_MISMATCH),
        (i64, VectorType(i64, [5]), SHAPE_MISMATCH),
        (VectorType(i32, [5]), VectorType(f32, [6]), SHAPE_MISMATCH),
        (VectorType(i32, [5]), VectorType(f64, [5]), BITWIDTH_MISMATCH),
        (MemRefType(i32, [5]), MemRefType(f32, [6]), SHAPE_MISMATCH),
        (MemRefType(i32, [5]), f32, SHAPE_MISMATCH),
    ],
)
def test_bitcast_incorrect(in_type: Attribute, out_type: Attribute, err_msg: str):
    in_arg = create_ssa_value(in_type)
    cast = BitcastOp(in_arg, out_type)

    with pytest.raises(VerifyException, match=err_msg):
        cast.verify_()


def test_index_cast_op():
    a = ConstantOp.from_int_and_width(0, 32)
    cast = IndexCastOp(a, IndexType())

    assert cast.result.type == IndexType()
    assert cast.input.type == i32
    assert cast.input.owner == a


def test_cast_fp_and_si_ops():
    a = ConstantOp.from_int_and_width(0, 32)
    fp = SIToFPOp(a, f32)
    si = FPToSIOp(fp, i32)

    assert fp.input == a.result
    assert fp.result == si.input
    assert isinstance(si.result.type, IntegerType)
    assert fp.result.type == f32


def test_cast_fp_and_ui_ops():
    a = ConstantOp.from_int_and_width(0, 32)
    fp = UIToFPOp(a, f32)
    ui = FPToUIOp(fp, i32)

    assert fp.input == a.result
    assert fp.result == ui.input
    assert isinstance(ui.result.type, IntegerType)
    assert fp.result.type == f32


def test_negf_op():
    a = ConstantOp(FloatAttr(1.0, f32))
    neg_a = NegfOp(a)

    b = ConstantOp(FloatAttr(1.0, f64))
    neg_b = NegfOp(b)

    assert neg_a.result.type == f32
    assert neg_b.result.type == f64


def test_extend_truncate_fpops():
    a = ConstantOp(FloatAttr(1.0, f32))
    b = ConstantOp(FloatAttr(2.0, f64))
    ext_op = ExtFOp(a, f64)
    trunc_op = TruncFOp(b, f32)

    assert ext_op.input == a.result
    assert ext_op.result.type == f64
    assert trunc_op.input == b.result
    assert trunc_op.result.type == f32


def test_cmpf_from_mnemonic():
    a = ConstantOp(FloatAttr(1.0, f64))
    b = ConstantOp(FloatAttr(2.0, f64))
    operations = [
        "false",
        "oeq",
        "ogt",
        "oge",
        "olt",
        "ole",
        "one",
        "ord",
        "ueq",
        "ugt",
        "uge",
        "ult",
        "ule",
        "une",
        "uno",
        "true",
    ]
    cmpf_ops = [CmpfOp(a, b, operations[i]) for i in range(len(operations))]

    for index, op in enumerate(cmpf_ops):
        assert op.lhs.type == f64
        assert op.rhs.type == f64
        assert op.predicate.value.data == index


def test_cmpf_get():
    a = ConstantOp(FloatAttr(1.0, f32))
    b = ConstantOp(FloatAttr(2.0, f32))

    cmpf_op = CmpfOp(a, b, 1)

    assert cmpf_op.lhs.type == f32
    assert cmpf_op.rhs.type == f32
    assert cmpf_op.predicate.value.data == 1


def test_cmpf_mismatch_type():
    a = ConstantOp(FloatAttr(1.0, f32))
    b = ConstantOp(FloatAttr(2.0, f64))

    with pytest.raises(
        TypeError,
        match="Comparison operands must have same type, but provided f32 and f64",
    ):
        _cmpf_op = CmpfOp(a, b, 1)


def test_cmpi_mismatch_type():
    a = ConstantOp.from_int_and_width(1, i32)
    b = ConstantOp.from_int_and_width(2, i64)

    with pytest.raises(
        TypeError,
        match="Comparison operands must have same type, but provided i32 and i64",
    ):
        _cmpi_op = CmpiOp(a, b, 1)


def test_cmpf_incorrect_comparison():
    a = ConstantOp(FloatAttr(1.0, f32))
    b = ConstantOp(FloatAttr(2.0, f32))

    with pytest.raises(VerifyException, match="Unknown comparison mnemonic: eq"):
        # 'eq' is a comparison op for cmpi but not cmpf
        _cmpf_op = CmpfOp(a, b, "eq")


def test_cmpi_incorrect_comparison():
    a = ConstantOp.from_int_and_width(1, i32)
    b = ConstantOp.from_int_and_width(2, i32)

    with pytest.raises(VerifyException, match="Unknown comparison mnemonic: oeq"):
        # 'oeq' is a comparison op for cmpf but not cmpi
        _cmpi_op = CmpiOp(a, b, "oeq")


def test_cmpi_index_type():
    a = ConstantOp.from_int_and_width(1, IndexType())
    b = ConstantOp.from_int_and_width(2, IndexType())
    CmpiOp(a, b, "eq").verify()


def test_extend_truncate_iops():
    a = ConstantOp.from_int_and_width(1, i32)
    b = ConstantOp.from_int_and_width(2, i64)
    exts_op = ExtSIOp(a, i64)
    extu_op = ExtUIOp(a, i64)
    trunc_op = TruncIOp(b, i32)
    exts_op.verify()
    extu_op.verify()
    trunc_op.verify()

    assert exts_op.input == a.result
    assert exts_op.result.type == i64
    assert extu_op.input == a.result
    assert extu_op.result.type == i64
    assert trunc_op.input == b.result
    assert trunc_op.result.type == i32


def test_trunci_incorrect_bitwidth():
    a = ConstantOp.from_int_and_width(1, 16)
    # bitwidth of b has to be smaller than the one of a
    with pytest.raises(VerifyException):
        _trunci_op = TruncIOp(a, i32).verify()


def test_extui_incorrect_bitwidth():
    a = ConstantOp.from_int_and_width(1, 64)
    # bitwidth of b has to be larger than the one of a
    with pytest.raises(VerifyException):
        _extui_op = ExtUIOp(a, i32).verify()



"""Dialect for unlimited precision integers with Python `int` semantics."""

import abc

from xdsl.dialects.builtin import f64, i1
from xdsl.ir import (
    Dialect,
    Operation,
    ParametrizedAttribute,
    SSAValue,
    TypeAttribute,
)
from xdsl.irdl import (
    IRDLOperation,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    result_def,
    traits_def,
)
from xdsl.traits import Commutative, Pure, SameOperandsAndResultType


@irdl_attr_definition
class BigIntegerType(ParametrizedAttribute, TypeAttribute):
    """Type for unlimited precision integers, with Python `int` semantics."""

    name = "bigint.bigint"


bigint = BigIntegerType()


class BinaryOperation(IRDLOperation, abc.ABC):
    """Binary operation where all operands and results are `bigint`s."""

    lhs = operand_def(bigint)
    rhs = operand_def(bigint)
    result = result_def(bigint)

    assembly_format = "$lhs `,` $rhs attr-dict `:` type($result)"

    def __init__(
        self,
        operand1: Operation | SSAValue,
        operand2: Operation | SSAValue,
    ):
        super().__init__(operands=[operand1, operand2], result_types=[bigint])


@irdl_op_definition
class AddOp(BinaryOperation):
    """Add two `bigint`s."""

    name = "bigint.add"

    traits = traits_def(
        Pure(),
        Commutative(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class SubOp(BinaryOperation):
    """Subtract two `bigint`s."""

    name = "bigint.sub"

    traits = traits_def(
        Pure(),
        Commutative(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class MulOp(BinaryOperation):
    """Multiply two `bigint`s."""

    name = "bigint.mul"

    traits = traits_def(
        Pure(),
        Commutative(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class FloorDivOp(BinaryOperation):
    """Floor divide two `bigint`s, rounding down to the nearest integer."""

    name = "bigint.floordiv"

    traits = traits_def(
        Pure(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class ModOp(BinaryOperation):
    """Modulo two `bigint`s, taking the sign of the divisor."""

    name = "bigint.mod"

    traits = traits_def(
        Pure(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class PowOp(BinaryOperation):
    """Exponentiate a `bigint` by another."""

    name = "bigint.pow"

    traits = traits_def(
        Pure(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class LShiftOp(BinaryOperation):
    """Left shift a `bigint` by another."""

    name = "bigint.lshift"

    traits = traits_def(
        Pure(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class RShiftOp(BinaryOperation):
    """Right shift a `bigint` by another."""

    name = "bigint.rshift"

    traits = traits_def(
        Pure(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class BitOrOp(BinaryOperation):
    """Bitwise OR a `bigint` with another."""

    name = "bigint.bitor"

    traits = traits_def(
        Pure(),
        Commutative(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class BitXorOp(BinaryOperation):
    """Bitwise XOR a `bigint` with another."""

    name = "bigint.bitxor"

    traits = traits_def(
        Pure(),
        Commutative(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class BitAndOp(BinaryOperation):
    """Bitwise AND a `bigint` with another."""

    name = "bigint.bitand"

    traits = traits_def(
        Pure(),
        Commutative(),
        SameOperandsAndResultType(),
    )


@irdl_op_definition
class DivOp(IRDLOperation):
    """Divide two `bigint`s, yielding a 64-bit floating point type.

    Note that this operation follows Python semantics, for example by rounding
    to minus infinity.
    """

    name = "bigint.div"

    lhs = operand_def(bigint)
    rhs = operand_def(bigint)
    result = result_def(f64)

    assembly_format = "$lhs `,` $rhs attr-dict `:` type($result)"

    traits = traits_def(
        Pure(),
    )

    def __init__(
        self,
        operand1: Operation | SSAValue,
        operand2: Operation | SSAValue,
    ):
        super().__init__(operands=[operand1, operand2], result_types=[f64])


class ComparisonOperation(IRDLOperation, abc.ABC):
    """Binary operation comparing two `bigint`s and returning a boolean."""

    lhs = operand_def(bigint)
    rhs = operand_def(bigint)
    result = result_def(i1)

    assembly_format = "$lhs `,` $rhs attr-dict `:` type($result)"

    def __init__(
        self,
        operand1: Operation | SSAValue,
        operand2: Operation | SSAValue,
    ):
        super().__init__(operands=[operand1, operand2], result_types=[i1])


@irdl_op_definition
class EqOp(ComparisonOperation):
    """Check equality of two `bigint`s."""

    name = "bigint.eq"

    traits = traits_def(
        Pure(),
        Commutative(),
    )


@irdl_op_definition
class NeqOp(ComparisonOperation):
    """Check inequality of two `bigint`s."""

    name = "bigint.neq"

    traits = traits_def(
        Pure(),
        Commutative(),
    )


@irdl_op_definition
class GtOp(ComparisonOperation):
    """Check if one `bigint` is greater than another."""

    name = "bigint.gt"

    traits = traits_def(
        Pure(),
    )


@irdl_op_definition
class GteOp(ComparisonOperation):
    """Check if one `bigint` is greater than or equal to another."""

    name = "bigint.gte"

    traits = traits_def(
        Pure(),
    )


@irdl_op_definition
class LtOp(ComparisonOperation):
    """Check if one `bigint` is less than another."""

    name = "bigint.lt"

    traits = traits_def(
        Pure(),
    )


@irdl_op_definition
class LteOp(ComparisonOperation):
    """Check if one `bigint` is less than or equal to another."""

    name = "bigint.lte"

    traits = traits_def(
        Pure(),
    )


BigInt = Dialect(
    "bigint",
    [
        AddOp,
        SubOp,
        MulOp,
        FloorDivOp,
        ModOp,
        PowOp,
        LShiftOp,
        RShiftOp,
        BitOrOp,
        BitXorOp,
        BitAndOp,
        DivOp,
        EqOp,
        NeqOp,
        GtOp,
        GteOp,
        LtOp,
        LteOp,
    ],
    [
        BigIntegerType,
    ],
)



from xdsl.dialects import arith, builtin
from xdsl.dialects.builtin import BoolAttr, IndexType, IntegerType
from xdsl.ir import OpResult
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.traits import Commutative
from xdsl.transforms.canonicalization_patterns.utils import (
    const_evaluate_operand,
    const_evaluate_operand_attribute,
)
from xdsl.utils.hints import isa


class SignlessIntegerBinaryOperationZeroOrUnitRight(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.SignlessIntegerBinaryOperation, rewriter: PatternRewriter, /
    ):
        if (rhs := const_evaluate_operand_attribute(op.rhs)) is None:
            return
        if op.is_right_zero(rhs):
            rewriter.replace_matched_op((), (op.rhs,))
        elif op.is_right_unit(rhs):
            rewriter.replace_matched_op((), (op.lhs,))


class SignlessIntegerBinaryOperationConstantProp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.SignlessIntegerBinaryOperation, rewriter: PatternRewriter, /
    ):
        if (lhs := const_evaluate_operand(op.lhs)) is None:
            return
        if (rhs := const_evaluate_operand(op.rhs)) is None:
            # Swap inputs if lhs is constant and rhs is not
            if op.has_trait(Commutative):
                rewriter.replace_matched_op(op.__class__(op.rhs, op.lhs))
            return

        if (res := op.py_operation(lhs, rhs)) is None:
            return
        assert isinstance(op.result.type, IntegerType | IndexType)

        rewriter.replace_matched_op(
            arith.ConstantOp.from_int_and_width(res, op.result.type, truncate_bits=True)
        )


def _fold_const_operation(
    op_t: type[arith.FloatingPointLikeBinaryOperation],
    lhs: builtin.FloatAttr,
    rhs: builtin.FloatAttr,
) -> arith.ConstantOp | None:
    match op_t:
        case arith.AddfOp:
            val = lhs.value.data + rhs.value.data
        case arith.SubfOp:
            val = lhs.value.data - rhs.value.data
        case arith.MulfOp:
            val = lhs.value.data * rhs.value.data
        case arith.DivfOp:
            if rhs.value.data == 0.0:
                # this mirrors what mlir does
                if lhs.value.data == 0.0:
                    val = float("nan")
                elif lhs.value.data < 0:
                    val = float("-inf")
                else:
                    val = float("inf")
            else:
                val = lhs.value.data / rhs.value.data
        case _:
            return
    return arith.ConstantOp(builtin.FloatAttr(val, lhs.type))


class FoldConstConstOp(RewritePattern):
    """
    Folds a floating point binary op whose operands are both `arith.constant`s.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.FloatingPointLikeBinaryOperation, rewriter: PatternRewriter, /
    ):
        if (
            isinstance(op.lhs.owner, arith.ConstantOp)
            and isinstance(op.rhs.owner, arith.ConstantOp)
            and isa(l := op.lhs.owner.value, builtin.FloatAttr)
            and isa(r := op.rhs.owner.value, builtin.FloatAttr)
            and (cnst := _fold_const_operation(type(op), l, r))
        ):
            rewriter.replace_matched_op(cnst)


class FoldConstsByReassociation(RewritePattern):
    """
    Rewrites a chain of
        `(const1 <op> var) <op> const2`
    as
        `folded_const <op> val`

    The op must be associative and have the `fastmath<reassoc>` flag set.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: arith.AddfOp | arith.MulfOp, rewriter: PatternRewriter, /
    ):
        if isinstance(op.lhs.owner, arith.ConstantOp):
            const1, val = op.lhs.owner, op.rhs
        else:
            const1, val = op.rhs.owner, op.lhs

        if (
            not isinstance(const1, arith.ConstantOp)
            or not isinstance(u := op.result.get_user_of_unique_use(), type(op))
            or not isinstance(
                const2 := u.lhs.owner if u.rhs == op.result else u.rhs.owner,
                arith.ConstantOp,
            )
            or arith.FastMathFlag.REASSOC not in op.fastmath.flags
            or arith.FastMathFlag.REASSOC not in u.fastmath.flags
            or not isa(c1 := const1.value, builtin.FloatAttr)
            or not isa(c2 := const2.value, builtin.FloatAttr)
        ):
            return

        if cnsts := _fold_const_operation(type(op), c1, c2):
            flags = arith.FastMathFlagsAttr(list(op.fastmath.flags | u.fastmath.flags))
            rebuild = type(op)(cnsts, val, flags)
            rewriter.replace_matched_op([cnsts, rebuild])
            rewriter.replace_op(u, [], [rebuild.result])


class SelectConstPattern(RewritePattern):
    """
    arith.select %true %x %y = %x
    arith.select %false %x %y = %y
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.SelectOp, rewriter: PatternRewriter):
        const_value = const_evaluate_operand(op.cond)

        if const_value is None:
            return

        new_results = (op.lhs,) if const_value else (op.rhs,)
        rewriter.replace_matched_op((), new_results)


class SelectTrueFalsePattern(RewritePattern):
    """
    arith.select %x %true %false = %x
    arith.select %x %false %true = %x xor 1
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.SelectOp, rewriter: PatternRewriter):
        if op.result.type != IntegerType(1):
            return

        if (lhs := const_evaluate_operand(op.lhs)) is None or (
            rhs := const_evaluate_operand(op.rhs)
        ) is None:
            return

        if lhs and not rhs:
            rewriter.replace_matched_op((), (op.cond,))

        if not lhs and rhs:
            rewriter.replace_matched_op(arith.XOrIOp(op.cond, op.rhs))


class SelectSamePattern(RewritePattern):
    """
    arith.select %x %y %y = %y
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.SelectOp, rewriter: PatternRewriter):
        if op.lhs == op.rhs:
            rewriter.replace_matched_op((), (op.lhs,))


class SelectFoldCmpfPattern(RewritePattern):
    """
    %1 = arith.cmpf  ogt, %0, %cst fastmath<nnan> : f64
    %2 = arith.select %1, %0, %cst : f64
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.SelectOp, rewriter: PatternRewriter):
        if not isinstance(op.cond, OpResult) or not isinstance(
            cmpf := op.cond.op, arith.CmpfOp
        ):
            return
        if (
            arith.FastMathFlag.NO_NANS not in cmpf.fastmath.flags
            or arith.FastMathFlag.NO_SIGNED_ZEROS not in cmpf.fastmath.flags
        ):
            return
        if not (op.lhs == cmpf.lhs and op.rhs == cmpf.rhs):
            return

        target = None
        match cmpf.predicate.value.data:
            case 2 | 3 | 9 | 10:
                # ogt | oge | ugt | uge
                target = arith.MaximumfOp
            case 4 | 5 | 11 | 12:
                # olt | ole | ult | ule
                target = arith.MinimumfOp
            case _:
                return
        rewriter.replace_matched_op(target(op.lhs, op.rhs, cmpf.fastmath))


class ApplyCmpiPredicateToEqualOperands(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.CmpiOp, rewriter: PatternRewriter):
        if op.lhs != op.rhs:
            return
        val = op.predicate.value.data in (0, 3, 5, 7, 9)
        rewriter.replace_matched_op(arith.ConstantOp(BoolAttr.from_bool(val)))



from collections.abc import Callable

from xdsl.context import Context
from xdsl.dialects import transform
from xdsl.interpreter import (
    Interpreter,
    InterpreterFunctions,
    PythonValues,
    ReturnedValues,
    TerminatorValue,
    impl,
    impl_callable,
    impl_terminator,
    register_impls,
)
from xdsl.passes import ModulePass, PassPipeline


@register_impls
class TransformFunctions(InterpreterFunctions):
    ctx: Context
    passes: dict[str, Callable[[], type[ModulePass]]]

    def __init__(
        self, ctx: Context, available_passes: dict[str, Callable[[], type[ModulePass]]]
    ):
        self.ctx = ctx
        self.passes = available_passes

    @impl_callable(transform.NamedSequenceOp)
    def run_named_sequence_op(
        self,
        interpreter: Interpreter,
        op: transform.NamedSequenceOp,
        args: PythonValues,
    ) -> PythonValues:
        return interpreter.run_ssacfg_region(op.body, args, op.sym_name.data)

    @impl(transform.ApplyRegisteredPassOp)
    def run_apply_registered_pass_op(
        self,
        interpreter: Interpreter,
        op: transform.ApplyRegisteredPassOp,
        args: PythonValues,
    ) -> PythonValues:
        pass_name = op.pass_name.data
        pipeline = PassPipeline.parse_spec(self.passes, pass_name)
        pipeline.apply(self.ctx, args[0])
        return (args[0],)

    @impl_terminator(transform.YieldOp)
    def run_yield_op(
        self, interpreter: Interpreter, op: transform.YieldOp, args: PythonValues
    ) -> tuple[TerminatorValue, PythonValues]:
        return ReturnedValues(args), ()



import marimo

__generated_with = "0.13.6"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo
    from xdsl.utils import marimo as xmo
    from xdsl.ir import Block, Region
    from xdsl.builder import Builder, InsertPoint
    from xdsl.dialects import builtin, riscv, riscv_cf, riscv_func
    from xdsl.printer import Printer
    from xdsl.parser import Parser
    from xdsl.context import Context
    import difflib
    return (
        Block,
        Builder,
        Context,
        InsertPoint,
        Parser,
        Printer,
        builtin,
        difflib,
        mo,
        riscv,
        riscv_cf,
        riscv_func,
        xmo,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# RISC-V Dialects""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    ## RISC-V

    The [RISC-V](https://riscv.org/) instruction set is a small, extensible set of instructions that is gaining in popularity in research and industry.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.accordion(
        {
            "Click here for some resources on the RISC-V ISA and assembly format": (
                """
    * [Cheat Sheet](https://www.cl.cam.ac.uk/teaching/1617/ECAD+Arch/files/docs/RISCVGreenCardv8-20151013.pdf)
    * [Assembly Manual](https://github.com/riscv-non-isa/riscv-asm-manual/blob/main/src/asm-manual.adoc)
    * [ISA Manual](https://github.com/riscv/riscv-isa-manual)
    * [Detailed Instruction Definition](https://msyksphinz-self.github.io/riscv-isadoc/html/rvi.html)
    * [ABI](https://d3s.mff.cuni.cz/files/teaching/nswi200/202324/doc/riscv-abi.pdf)
    * [Formal Specification](https://github.com/riscv/sail-riscv)"""
            ),
        }
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Simple Arithmetic in RISC-V""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    Here is a simple C function and its corresponding assembly:

    ```C
    int fused_multiply_add(int a, int b, int c) {
        return a + b * c;
    }
    ```

    ```asm
    # Label corresponding to the function name
    # Arguments passed in a0, a1, a2 registers
    # Result expected to be stored in a0 register at the end of execution
    fused_multiply_add:
        # a1 <- a2 * a1
        mul     a1, a2, a1
        # a0 <- a0 + a1
        add     a0, a0, a1
        # Assembly pseudo-operation that jumps to the caller-passed return address
        ret
    ```

    ([Compiler Explorer](https://godbolt.org/z/E67a877vG))
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    A few aspects of assembly make it difficult to reason about directly.
    It doesn't track dependencies explicitly, making it difficult to know what the values in the inputs of an operation are, and whether its results will be used.
    Control flow is done via jumps to labels, such as the `fused_multiply_add` above.
    In order to represent and reason about assembly-level code, we introduce a set of backend dialects: `riscv`, `riscv_func`, and `riscv_cf`.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## The `riscv` Dialect""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    The `risv` dialect contains definitions for static information and operations.

    The four kinds of static information in RISC-V assembly are strings, integers, labels, and registers.
    The `riscv` dialect includes integer and float registers (`IntRegisterType` & `FloatRegisterType`).
    Both of these register classes encode both the information about the binary encoding of the register (e.g. `x0`, `x11`) and its pretty assembly name (e.g. `zero`, `a1`).

    The pretty assembly names reflect the RISC-V ABI. The registers `x0` to `x4` are specified to store some system information: `zero` for a register that always has the 0 value, `ra` to store the return address to jump to, `sp` for stack pointer, `gp` for global pointer, `tp` for thread pointer. The `a0-a7` registers are used to pass function arguments and results. The `s0-s11` registers are expected to have the same values before and after function calls. The `t0-t6` registers are temporary registers that are not expected to have the same values before and after function calls.
    """
    )
    return


@app.cell
def _(riscv):
    # Explictly constructing registers

    riscv.IntRegisterType.from_name("zero"), riscv.IntRegisterType.from_name("a1"), riscv.IntRegisterType.from_name("s2"), riscv.IntRegisterType.from_name("t3")
    return


@app.cell
def _(riscv):
    # Using the Registers helper

    riscv.Registers.ZERO, riscv.Registers.A1, riscv.Registers.S2, riscv.Registers.T3
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    We use these registers as the types of the operands and results of operations in the riscv dialect.
    Let's construct a dummy block with some operations in it:
    """
    )
    return


@app.cell
def _(Block, Builder, InsertPoint, Printer, riscv):
    block = Block(arg_types=(riscv.Registers.A0, riscv.Registers.A1))
    a0, a1 = block.args
    builder = Builder(InsertPoint.at_end(block))

    # Explicitly specify result registers
    addi_op = builder.insert(riscv.AddiOp(a0, 42, rd=riscv.Registers.A2))
    sub_op = builder.insert(riscv.SubOp(a0, a1, rd=riscv.Registers.A3))
    mul_op = builder.insert(riscv.MulOp(addi_op.rd, sub_op.rd, rd=riscv.Registers.A4))

    Printer().print_block(block)
    return (block,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    Note that the _i_ in the `addi` operation above stands for _immediate_, or a constant value in the assembly.
    These can be integers or labels, in which case the assembler will resolve the immediate to the value referred to by the label.

    Almost all the operations in the riscv dialect correspond to the instructions in the RISC-V ISA.
    RISC-V assembly has operations that correspond 1:1 with instructions, as well as pseudo-operations that are syntactic sugar for other operations, such as the `nop` instruction that desugars to `addi x0, x0, 0`.

    [The RISC-V Assembly Manual](https://github.com/riscv-non-isa/riscv-asm-manual/blob/main/src/asm-manual.adoc#a-listing-of-standard-risc-v-pseudoinstructions) has the list of pseudo-instructions in RISC-V.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Printing Assembly""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""All operations in the `riscv` dialect implement an `assembly_line` function that communicates how they should be displayed in assembly:""")
    return


@app.cell
def _(block):
    for op in block.ops:
        print(op.assembly_line())
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Functions and the `riscv_func` Dialect""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    When compiling a function for certain hardware, an Application Binary Interface (ABI) specifies things like which registers will hold the function arguments when the function begins, which registers will hold the results when the function ends, which registers are overridden, etc.

    We use the `riscv_func` dialect to represent RISC-V-specific functions that encode these calling convention considerations in the IR.

    Here is our `mul` function from above:
    """
    )
    return


@app.cell
def _():
    mul_ir = """
    riscv_func.func @mul(%num : !riscv.reg<a0>) -> !riscv.reg<a0> {
      %res = riscv.mul %num, %num : (!riscv.reg<a0>, !riscv.reg<a0>) -> !riscv.reg<a0>
      riscv_func.return %res : !riscv.reg<a0>
    }
    """
    return (mul_ir,)


@app.cell
def _(Parser, ctx, mul_ir, riscv, xmo):
    mul_module = Parser(ctx, mul_ir).parse_module()

    xmo.asm_html(riscv.riscv_code(mul_module))
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Unstructured Control Flow and the `riscv_cf` Dialect""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""The C programming language allows for both _structured_ control flow with the `if`, `for`, and `while` constructs, and _unstructured_ flow using goto. The two functions below lower to exactly the same assembly:""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    ```C
    int switch1(int a, int b, int c) {
        if (a) {
            return b;
        } else {
            return c;
        }
    }

    int switch2(int a, int b, int c) {
        if (!a) {
            goto c_label;
        }
        return b;
    c_label:
        return c;
    }
    ```

    ([Compiler Explorer](https://godbolt.org/z/Mzh4ojG3r))
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    At the assembly level, the code structure looks much more like the code in `switch2`.

    To represent jumps in the assembly, we use blocks and successors.
    For example, the [beq](https://msyksphinz-self.github.io/riscv-isadoc/html/rvi.html#beq) instruction, which jumps by (or _breaks to_) the specified offset (or to the specified label) if the values in the source registers are equal, is represented with the `riscv.BeqOp` operation:
    """
    )
    return


@app.cell(hide_code=True)
def _(Parser, ctx, xmo):
    switch_ir = """\
    riscv_func.func @switch(%a : !riscv.reg<a0>, %b : !riscv.reg<a1>, %c : !riscv.reg<a2>) -> !riscv.reg<a0> {
      %zero = riscv.get_register : !riscv.reg<zero>
      riscv_cf.beq %a : !riscv.reg<a0>, %zero : !riscv.reg<zero>, ^2(), ^1()
    ^1():
      %res_b = riscv.mv %b : (!riscv.reg<a1>) -> !riscv.reg<a0>
      riscv_func.return %res_b : !riscv.reg<a0>
    ^2():
      riscv.label "c_label"
      %res_c = riscv.mv %c : (!riscv.reg<a2>) -> !riscv.reg<a0>
      riscv_func.return %res_c : !riscv.reg<a0>
    }
    """
    switch_module = Parser(ctx, switch_ir).parse_module()
    switch_module.verify()
    xmo.module_html(switch_module)
    return (switch_module,)


@app.cell
def _(riscv, switch_module):
    # We can print this IR as assembly
    switch_asm = riscv.riscv_code(switch_module)
    print(switch_asm)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    Note that there are a few discrepancies between the assembly and the IR.
    While in the IR we specified both the "then" and the "else" successors, in assembly the block implicitly falls through.
    We represent this by adding a verification method to the jump operations that checks that the "else" block is the one immediately following the conditional jump.
    The second discrepancy is that the conditional jump in assembly specifies the label to jump to.
    During assembly printing, the `BeqOp` looks at the first operation in the block and, if it finds a `LabelOperation`, prints the corresponding value.
    """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Exercise: Fib IR""")
    return


@app.cell(hide_code=True)
def _():
    fib_text = """\
    # The label corresponding to the fib function
    fib:
        bge zero, a0, .LBB1_4
        li a2, 1
        li a3, 1
    # Label inserted by Clang corresponding to the loop body
    .LBB1_2:
        add a4, a2, a3
        addi a0, a0, -1
        mv a1, a3
        mv a2, a3
        mv a3, a4
        # Jump to beginning of loop body if need to do more work
        bne zero, a0, .LBB1_2
        # Moves the result value to the first integer return register, as specified by ABI
        mv a0, a1
        ret
    # Label inserted by Clang corresponding to the case where the input is <= 0, can just return 1
    .LBB1_4:
        li a0, 1
        ret
    """
    return (fib_text,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
    In this exercise, we will rewrite the following function to assembly-level IR:

    ```C
    // Assume n non-negative
    int fib(int n) {
        int a = 1;
        int b = 1;
        for (int i = 0; i < n; ++i) {
            int c = a + b;
            a = b;
            b = c;
        }
        return a;
    }
    ```
    """
    )
    return


@app.cell(hide_code=True)
def _(fib_text, mo):
    mo.md(
        fr"""
    This is the expected assembly:

    ```asm
    {fib_text}
    ```
    """
    )
    return


@app.cell(hide_code=True)
def _(comment_only_line, fib_text):
    fib_cleaned = "\n".join(line for line in fib_text.splitlines() if not comment_only_line.findall(line))
    return (fib_cleaned,)


@app.cell(hide_code=True)
def _(mo):
    fib_editor = mo.ui.code_editor("""\
    riscv_func.func @fib(%num : !riscv.reg<a0>) -> !riscv.reg<a0> {
      %zero = riscv.get_register : !riscv.reg<zero>
      riscv_cf.bge %zero: !riscv.reg<zero>, %num :!riscv.reg<a0>, ^4(), ^1()
    ^1():
      %a_init = riscv.li 1 : !riscv.reg<a2>
      %b_init = riscv.li 1 : !riscv.reg<a3>
      riscv_cf.branch ^2 (%num : !riscv.reg<a0>, %a_init : !riscv.reg<a2>, %b_init : !riscv.reg<a3>)
    ^2(%i : !riscv.reg<a0>, %a_in : !riscv.reg<a2>, %b_in : !riscv.reg<a3>):
      riscv.label ".LBB1_2"
      %sum = riscv.li 2 : !riscv.reg<a4>
      %i_next = riscv.li 3 : !riscv.reg<a0>
      %temp = riscv.li 4 : !riscv.reg<a1>
      %a_next = riscv.li 5 : !riscv.reg<a2>
      %b_next = riscv.li 6 : !riscv.reg<a3>
      riscv_cf.bne %zero: !riscv.reg<zero>, %i_next : !riscv.reg<a0>, ^2(%i_next : !riscv.reg<a0>, %a_next : !riscv.reg<a2>, %b_next : !riscv.reg<a3>), ^3()
    ^3():
      %res = riscv.mv %temp : (!riscv.reg<a1>) -> !riscv.reg<a0>
      riscv_func.return %num : !riscv.reg<a0>
    ^4():
      riscv.label ".LBB1_4"
      %res_early = riscv.li 1 : !riscv.reg<a0>
      riscv_func.return %res_early : !riscv.reg<a0>
    }""", language="javascript")
    return (fib_editor,)


@app.cell(hide_code=True)
def _(fib_editor, mo):
    mo.md(
        f"""
    Modify the following IR to reduce the diff below:

    {fib_editor}
    """
    )
    return


@app.cell(hide_code=True)
def _(Parser, ctx, difflib, fib_cleaned, fib_editor, mo, riscv):
    try:
        _module = Parser(ctx, fib_editor.value).parse_module()
        _module.verify()
        _asm = riscv.riscv_code(_module)
        _diff = "\n".join(difflib.ndiff(_asm.splitlines(), fib_cleaned.splitlines()))
        _res = f"""
    ```asm
    {_diff}
    ```
    """
    except Exception as e:
        _res = f"{type(e).__name__}: {e}"

    mo.md(f"""
    Diff:

    {_res}
    """)
    return


@app.cell(hide_code=True)
def _():
    # Solution

    _ = """\
    riscv_func.func @fib(%num : !riscv.reg<a0>) -> !riscv.reg<a0> {
      %zero = riscv.get_register : !riscv.reg<zero>
      riscv_cf.bge %zero: !riscv.reg<zero>, %num :!riscv.reg<a0>, ^4(), ^1()
    ^1():
      %a_init = riscv.li 1 : !riscv.reg<a2>
      %b_init = riscv.li 1 : !riscv.reg<a3>
      riscv_cf.branch ^2 (%num : !riscv.reg<a0>, %a_init : !riscv.reg<a2>, %b_init : !riscv.reg<a3>)
    ^2(%i : !riscv.reg<a0>, %a_in : !riscv.reg<a2>, %b_in : !riscv.reg<a3>):
      riscv.label ".LBB1_2"
      %sum = riscv.add %a_in, %b_in : (!riscv.reg<a2>, !riscv.reg<a3>) -> !riscv.reg<a4>
      %i_next = riscv.addi %i, -1 : (!riscv.reg<a0>) -> !riscv.reg<a0>
      %temp = riscv.mv %b_in : (!riscv.reg<a3>) -> !riscv.reg<a1>
      %a_next = riscv.mv %b_in : (!riscv.reg<a3>) -> !riscv.reg<a2>
      %b_next = riscv.mv %sum : (!riscv.reg<a4>) -> !riscv.reg<a3>
      riscv_cf.bne %zero: !riscv.reg<zero>, %i_next : !riscv.reg<a0>, ^2(%i_next : !riscv.reg<a0>, %a_next : !riscv.reg<a2>, %b_next : !riscv.reg<a3>), ^3()
    ^3():
      %res = riscv.mv %temp : (!riscv.reg<a1>) -> !riscv.reg<a0>
      riscv_func.return %num : !riscv.reg<a0>
    ^4():
      riscv.label ".LBB1_4"
      %res_early = riscv.li 1 : !riscv.reg<a0>
      riscv_func.return %res_early : !riscv.reg<a0>
    }
    """
    return


@app.cell(hide_code=True)
def _(Context, builtin, riscv, riscv_cf, riscv_func):
    ctx = Context()
    ctx.load_dialect(builtin.Builtin)
    ctx.load_dialect(riscv.RISCV)
    ctx.load_dialect(riscv_cf.RISCV_Cf)
    ctx.load_dialect(riscv_func.RISCV_Func)
    return (ctx,)


@app.cell(hide_code=True)
def _():
    import re

    comment_only_line = re.compile(r"^\s*#.*$")
    return (comment_only_line,)


if __name__ == "__main__":
    app.run()



import os
import typing
from dataclasses import dataclass, field
from importlib import import_module
from types import ModuleType
from typing import Any, cast

import xdsl.ir
import xdsl.irdl
from xdsl.ir import Attribute, Dialect, OpResult, ParametrizedAttribute, Region
from xdsl.irdl import (
    AllOf,
    AnyAttr,
    AnyOf,
    AttrConstraint,
    AttributeDef,
    BaseAttr,
    EqAttrConstraint,
    IRDLOperation,
    OperandDef,
    OptAttributeDef,
    OptOperandDef,
    OptPropertyDef,
    OptRegionDef,
    OptResultDef,
    OptSuccessorDef,
    ParamAttrConstraint,
    PropertyDef,
    RangeOf,
    RegionDef,
    ResultDef,
    SuccessorDef,
    VarOperandDef,
    VarRegionDef,
    VarResultDef,
    VarSuccessorDef,
)


@dataclass
class DialectStubGenerator:
    """Generate a typing stub file (.pyi) for a dialect."""

    dialect: Dialect
    dependencies: dict[str, set[str]] = field(
        init=False, default_factory=dict[str, set[str]]
    )

    def _import(self, module: ModuleType | str, name: str | type[Any]):
        """
        Internal helper to keep track of dependencies, to later generate clean import
        statements.
        """
        # If passed a type, use its name.
        if isinstance(name, type):
            name = name.__name__
        # If passed a module, use its name.
        if isinstance(module, ModuleType):
            module = module.__name__
        # Do not import from builtins, those are the implicitely available ones.
        if module == "builtins":
            return
        # Do no import from banned nested modules
        if module.startswith("xdsl.ir."):
            module = "xdsl.ir"
        if module.startswith("xdsl.irdl."):
            module = "xdsl.irdl"
        # Create a module dependency or add a new name to the ones from an existing
        # dependency
        if module in self.dependencies:
            self.dependencies[module].add(name)
        else:
            self.dependencies[module] = {name}

    def _generate_constraint_type(self, constraint: AttrConstraint) -> str:
        """
        Return a type hint for the member constrained by a constraint, by it an
        attribute parameter, or an operation attribute/property.
        """
        import xdsl.dialects.builtin
        import xdsl.ir
        from xdsl.dialects.builtin import ArrayAttr, ArrayOfConstraint

        match constraint:
            case BaseAttr(attr_type):
                if attr_type not in self.dialect.attributes:
                    self._import(attr_type.__module__, attr_type.__name__)
                return attr_type.__name__
            case EqAttrConstraint(attr):
                if type(attr) not in self.dialect.attributes:
                    self._import(type(attr).__module__, type(attr).__name__)
                return type(attr).__name__

            case AnyOf(attr_constrs=constraints):
                return " | ".join(
                    self._generate_constraint_type(c) for c in constraints
                )
            case AllOf(constraints):
                self._import(typing, "Annotated")
                return f"Annotated[{', '.join(self._generate_constraint_type(c) for c in reversed(constraints))}]"  # noqa: E501
            case ArrayOfConstraint(RangeOf(constraint)):
                self._import(xdsl.dialects.builtin, ArrayAttr)
                return f"ArrayAttr[{self._generate_constraint_type(constraint)}]"
            case AnyAttr():
                self._import(xdsl.ir, Attribute)
                return "Attribute"
            case ParamAttrConstraint():
                base_type = cast(
                    ParamAttrConstraint[ParametrizedAttribute], constraint
                ).base_attr
                return base_type.__name__

            case _:
                raise NotImplementedError(
                    f"Unsupported constraint type: {type(constraint)}"
                )

    def _generate_attribute_stub(self, attr: type[ParametrizedAttribute]):
        """
        Generate type stub for an irdl attribute.
        """
        # They all are ParametrizedAttributes.
        self._import(xdsl.ir, ParametrizedAttribute)
        # Get the bases that are not already bases of ParametrizedAttribute.
        bases = set(attr.__mro__[1:]) - set(ParametrizedAttribute.__mro__)
        # Add them as stub dependencies
        for base in bases:
            self._import(base.__module__, base)

        # Also add them to the Attribute class' bases.
        bases = ", ".join(b.__name__ for b in bases)
        if bases:
            bases += ", "
        yield f"class {attr.__name__}({bases}ParametrizedAttribute):"

        # Generate the parameters' stubs, if any
        attr_def = attr.get_irdl_definition()
        for name, param in attr_def.parameters:
            yield f'    {name} : "{self._generate_constraint_type(param.constr)}"'
        # Otherwise, generate a pass for Python's indentation
        if not attr_def.parameters:
            yield "    pass"
        yield ""
        yield ""

    def _generate_operation_stub(self, op: type[IRDLOperation]):
        """
        Generate type stub for an irdl operation.
        """
        # Keep track of whether the operation has any body, to generate a pass if it
        # does not.
        had_body = False

        # They all are IRDLOperations.
        self._import(xdsl.irdl, IRDLOperation)

        # Currently there's nothing that should be a base class or the IRDLOperations.
        # Traits are not supported, and implemented as fields in PyRDL.
        yield f"class {op.__name__}(IRDLOperation):"

        # Generate the constructs' stubs, if any
        op_def = op.get_irdl_definition()
        for name, o in op_def.operands:
            had_body = True
            match o:
                case VarOperandDef(_):
                    self._import(xdsl.irdl, "VarOperand")
                    yield f"    {name} : VarOperand"
                case OptOperandDef(_):
                    self._import(xdsl.irdl, "OptOperand")
                    yield f"    {name} : OptOperand"
                case OperandDef(_):
                    self._import(xdsl.irdl, "Operand")
                    yield f"    {name} : Operand"
        for name, o in op_def.results:
            had_body = True
            match o:
                case VarResultDef():
                    self._import(xdsl.irdl, "VarOpResult")
                    yield f"    {name} : VarOpResult"
                case OptResultDef():
                    self._import(xdsl.irdl, "OptOpResult")
                    yield f"    {name} : OptOpResult"
                case ResultDef():
                    self._import(xdsl.ir, OpResult)
                    yield f"    {name} : OpResult"
        for name, o in op_def.attributes.items():
            had_body = True
            match o:
                case OptAttributeDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)} | None"  # noqa: E501
                case AttributeDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)}"
        for name, o in op_def.properties.items():
            had_body = True
            match o:
                case OptPropertyDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)} | None"  # noqa: E501
                case PropertyDef():
                    yield f"    {name} : {self._generate_constraint_type(o.constr)}"

        for name, r in op_def.regions:
            had_body = True
            match r:
                case OptRegionDef():
                    self._import(xdsl.irdl, "OptRegion")
                    yield f"    {name} : OptRegion"
                case VarRegionDef():
                    self._import(xdsl.irdl, "VarRegion")
                    yield f"    {name} : VarRegion"
                case RegionDef():
                    self._import(xdsl.ir, Region)
                    yield f"    {name} : Region"

        for name, r in op_def.successors:
            had_body = True
            match r:
                case OptSuccessorDef():
                    self._import(xdsl.irdl, "OptSuccessor")
                    yield f"    {name} : OptSuccessor"
                case VarSuccessorDef():
                    self._import(xdsl.irdl, "VarSuccessor")
                    yield f"    {name} : VarSuccessor"
                case SuccessorDef():
                    self._import(xdsl.irdl, "Successor")
                    yield f"    {name} : Successor"
        # Generate a pass if the operation had no body.
        if not had_body:
            yield "    pass"
        yield ""
        yield ""

    def _generate_dialect_stubs(self):
        """
        Generate a dialect's stubs.

        Just generate stubs for all attributes and operations in the dialect.
        """
        for attr in self.dialect.attributes:
            if issubclass(attr, ParametrizedAttribute):
                for l in self._generate_attribute_stub(attr):
                    yield l

        for op in self.dialect.operations:
            if issubclass(op, IRDLOperation):
                for l in self._generate_operation_stub(op):
                    yield l

    def _generate_imports(self):
        """
        Generate import statements for all the dependencies of the stub.
        """
        # sort modules alphabetically for deterministic and clean output.
        items = list(self.dependencies.items())
        items.sort()

        for module, names in items:
            # If only one name is imported from a module, make a one-liner import.
            if len(names) == 1:
                name = names.pop()
                yield f"from {module} import {name}"
            # Otherwise, import all names in a multi-line import, sorted again for
            # a deterministic and clean output.
            else:
                names = list(names)
                names.sort()
                yield f"from {module} import ("
                for o in names:
                    yield f"    {o},"
                yield ")"

    def generate_dialect_stubs(self):
        """
        The main function, generate stubs for the passed dialect and return as a string.

        NB: probably not optimal perf-wise, but I don't foresee this as a bottleneck.
        """
        self._import(xdsl.ir, Dialect)
        dialect_body = "\n".join(self._generate_dialect_stubs())
        imports = "\n".join(self._generate_imports())
        if imports:
            imports += "\n"

        return f"""\
{imports}
{dialect_body}
{self.dialect.name.capitalize()} : Dialect
"""


def make_all_stubs():
    import xdsl.dialects

    dialects = xdsl.dialects
    directory = "/".join(dialects.__path__)
    for file in os.listdir(directory):
        name, ext = os.path.splitext(file)
        if ext == ".irdl":
            import_module(f"{directory}/{name}")


if __name__ == "__main__":
    make_all_stubs()



from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass, field

from xdsl.ir import Attribute, Dialect, Operation, TypeAttribute
from xdsl.utils.exceptions import (
    AlreadyRegisteredConstructException,
    UnregisteredConstructException,
)


@dataclass
class Context:
    """Contains structures for operations/attributes registration."""

    allow_unregistered: bool = field(default=False)

    _loaded_dialects: dict[str, Dialect] = field(default_factory=dict[str, Dialect])
    _loaded_ops: dict[str, type[Operation]] = field(
        default_factory=dict[str, type[Operation]]
    )
    _loaded_attrs: dict[str, type[Attribute]] = field(
        default_factory=dict[str, type[Attribute]]
    )
    _loaded_types: dict[str, type[TypeAttribute]] = field(
        default_factory=dict[str, type[TypeAttribute]]
    )
    _registered_dialects: dict[str, Callable[[], Dialect]] = field(
        default_factory=dict[str, Callable[[], Dialect]]
    )
    """
    A dictionary of all registered dialects that are not yet loaded. This is used to
    only load the respective Python files when the dialect is actually used.
    """

    def clone(self) -> "Context":
        return Context(
            self.allow_unregistered,
            self._loaded_dialects.copy(),
            self._loaded_ops.copy(),
            self._loaded_attrs.copy(),
            self._loaded_types.copy(),
            self._registered_dialects.copy(),
        )

    @property
    def loaded_ops(self) -> "Iterable[type[Operation]]":
        """
        Returns all the loaded operations. Not valid across mutations of this object.
        """
        return self._loaded_ops.values()

    @property
    def loaded_attrs(self) -> "Iterable[type[Attribute]]":
        """
        Returns all the loaded attributes. Not valid across mutations of this object.
        """
        return self._loaded_attrs.values()

    @property
    def loaded_types(self) -> Iterable[type[TypeAttribute]]:
        """
        Returns all the loaded types. Not valid across mutations of this object.
        """
        return self._loaded_types.values()

    @property
    def loaded_dialects(self) -> "Iterable[Dialect]":
        """
        Returns all the loaded attributes. Not valid across mutations of this object.
        """
        return self._loaded_dialects.values()

    @property
    def registered_dialect_names(self) -> Iterable[str]:
        """
        Returns the names of all registered dialects. Not valid across mutations of this object.
        """
        return self._registered_dialects.keys()

    def register_dialect(
        self, name: str, dialect_factory: "Callable[[], Dialect]"
    ) -> None:
        """
        Register a dialect without loading it. The dialect is only loaded in the context
        when an operation or attribute of that dialect is parsed, or when explicitely
        requested with `load_registered_dialect`.
        """
        if name in self._registered_dialects:
            raise AlreadyRegisteredConstructException(
                f"'{name}' dialect is already registered"
            )
        self._registered_dialects[name] = dialect_factory

    def load_registered_dialect(self, name: str) -> None:
        """Load a dialect that is already registered in the context."""
        if name not in self._registered_dialects:
            raise UnregisteredConstructException(f"'{name}' dialect is not registered")
        dialect = self._registered_dialects[name]()
        self._loaded_dialects[dialect.name] = dialect

        for op in dialect.operations:
            self.load_op(op)

        for attr in dialect.attributes:
            self.load_attr_or_type(attr)

    def load_dialect(self, dialect: "Dialect"):
        """
        Load a dialect. Operation and Attribute names should be unique.
        If the dialect is already registered in the context, use
        `load_registered_dialect` instead.
        """
        if dialect.name in self._registered_dialects:
            raise AlreadyRegisteredConstructException(
                f"'{dialect.name}' dialect is already registered, use 'load_registered_dialect' instead"
            )
        self.register_dialect(dialect.name, lambda: dialect)
        self.load_registered_dialect(dialect.name)

    def load_op(self, op: "type[Operation]") -> None:
        """Load an operation definition. Operation names should be unique."""
        if op.name in self._loaded_ops:
            raise AlreadyRegisteredConstructException(
                f"Operation {op.name} has already been loaded"
            )
        self._loaded_ops[op.name] = op

    def load_attr_or_type(self, attr: type[Attribute]) -> None:
        """
        Load an attribute or type definition.
        Attribute or type names should be unique, but an attribute may have the same
        name as a type.
        """
        if issubclass(attr, TypeAttribute):
            if attr.name in self._loaded_types:
                raise AlreadyRegisteredConstructException(
                    f"Type {attr.name} has already been loaded"
                )
            self._loaded_types[attr.name] = attr
        else:
            if attr.name in self._loaded_attrs:
                raise AlreadyRegisteredConstructException(
                    f"Attribute {attr.name} has already been loaded"
                )
            self._loaded_attrs[attr.name] = attr

    def _get_known_op(self, name: str) -> "type[Operation] | None":
        if name in self._loaded_ops:
            return self._loaded_ops[name]
        if "." in name:
            dialect_name, _ = Dialect.split_name(name)
            if (
                dialect_name in self._registered_dialects
                and dialect_name not in self._loaded_dialects
            ):
                self.load_registered_dialect(dialect_name)
                return self._get_known_op(name)

    def get_optional_op(
        self, name: str, *, dialect_stack: Sequence[str] = ()
    ) -> "type[Operation] | None":
        """
        Get an operation class from its name if it exists or is contained in one of the
        dialects in the dialect stack.
        If the operation is not registered, return None unless unregistered operations
        are allowed in the context, in which case return an UnregisteredOp.
        """
        # Check if the name is known.
        if op_type := self._get_known_op(name):
            return op_type

        # Check appending each dialect in the dialect stack.
        for dialect_name in reversed(dialect_stack):
            dialect_and_name = f"{dialect_name}.{name}"
            if op_type := self._get_known_op(dialect_and_name):
                return op_type

        # If the context allows unregistered operations then create an UnregisteredOp
        if self.allow_unregistered:
            from xdsl.dialects.builtin import UnregisteredOp

            op_type = UnregisteredOp.with_name(name)
            self._loaded_ops[name] = op_type
            return op_type

    def get_op(
        self, name: str, *, dialect_stack: Sequence[str] = ()
    ) -> "type[Operation]":
        """
        Get an operation class from its name if it exists or is contained in one of the
        dialects in the dialect stack.
        If the operation is not registered, raise an exception unless unregistered
        operations are allowed in the context, in which case return an UnregisteredOp.
        """
        if op_type := self.get_optional_op(name, dialect_stack=dialect_stack):
            return op_type
        raise UnregisteredConstructException(f"Operation {name} is not registered")

    def get_optional_type(self, name: str) -> type[TypeAttribute] | None:
        """
        Get a type definition from its name if it exists.
        If the type is not registered, return None unless unregistered types
        are allowed in the context, in which case return an UnregisteredAttr.
        """
        # If the type is already loaded, returns it.
        if name in self._loaded_types:
            return self._loaded_types[name]

        # Otherwise, check if the type dialect is registered.
        dialect_name, _ = Dialect.split_name(name)
        if (dialect_name in self._registered_dialects) and (
            dialect_name not in self._loaded_dialects
        ):
            self.load_registered_dialect(dialect_name)
            return self.get_optional_type(name)

        # If the dialect is unregistered, but the context allows unregistered
        # dialects, return an UnregisteredAttr.
        if self.allow_unregistered:
            from xdsl.dialects.builtin import UnregisteredAttr

            attr_type = UnregisteredAttr.with_name_and_type(name, True)
            assert issubclass(attr_type, TypeAttribute)
            self._loaded_types[name] = attr_type
            return attr_type

    def get_type(self, name: str) -> type[TypeAttribute]:
        """
        Get a type definition from its name.
        If the type is not registered, raise an exception unless unregistered
        types are allowed in the context, in which case return an UnregisteredAttr.
        """
        if attr_type := self.get_optional_type(name):
            return attr_type
        raise UnregisteredConstructException(f"Type {name} is not registered")

    def get_optional_attr(
        self,
        name: str,
    ) -> "type[Attribute] | None":
        """
        Get an attribute class from its name if it exists.
        If the attribute is not registered, return None unless unregistered attributes
        are allowed in the context, in which case return an UnregisteredAttr.
        """
        # If the attribute is already loaded, returns it.
        if name in self._loaded_attrs:
            return self._loaded_attrs[name]

        # Otherwise, check if the attribute dialect is registered.
        dialect_name, _ = Dialect.split_name(name)
        if (
            dialect_name in self._registered_dialects
            and dialect_name not in self._loaded_dialects
        ):
            self.load_registered_dialect(dialect_name)
            return self.get_optional_attr(name)

        # If the dialect is unregistered, but the context allows unregistered
        # attributes, return an UnregisteredOp.
        if self.allow_unregistered:
            from xdsl.dialects.builtin import UnregisteredAttr

            attr_type = UnregisteredAttr.with_name_and_type(name, False)
            self._loaded_attrs[name] = attr_type
            return attr_type

        return None

    def get_attr(
        self,
        name: str,
    ) -> "type[Attribute]":
        """
        Get an attribute class from its name.
        If the attribute is not registered, raise an exception unless unregistered
        attributes are allowed in the context, in which case return an UnregisteredAttr.
        """
        if attr_type := self.get_optional_attr(name):
            return attr_type
        raise UnregisteredConstructException(f"Attribute {name} is not registered")

    def get_dialect(self, name: str) -> "Dialect":
        if (dialect := self.get_optional_dialect(name)) is None:
            raise UnregisteredConstructException(f"Dialect {name} is not registered")
        return dialect

    def get_optional_dialect(self, name: str) -> "Dialect | None":
        """
        Get a dialect from its name if it exists.
        If the dialect is not registered, return None.
        """
        if name in self._registered_dialects and name not in self._loaded_dialects:
            self.load_registered_dialect(name)

        if name in self._loaded_dialects:
            return self._loaded_dialects[name]

        return None



from __future__ import annotations

from abc import ABC
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Any, TypeGuard, cast

from immutabledict import immutabledict

from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    Operation,
    OpResult,
    Region,
    SSAValue,
)
from xdsl.utils.exceptions import InvalidIRException
from xdsl.utils.immutable_list import IList


@dataclass(frozen=True)
class ISSAValue(ABC):
    """
    Represents an immutable SSA variable. An immutable SSA variable is either an operation result
    or a basic block argument.
    """

    type: Attribute
    users: IList[IOperation]

    def _add_user(self, op: IOperation):
        self.users._unfreeze()  # pyright: ignore[reportPrivateUsage]
        self.users.append(op)
        self.users.freeze()

    def _remove_user(self, op: IOperation):
        if op not in self.users:
            raise Exception(
                f"Trying to remove a user ({op.name}) that is not an actual user of this value!"
            )

        self.users._unfreeze()  # pyright: ignore[reportPrivateUsage]
        self.users.remove(op)
        self.users.freeze()


@dataclass(frozen=True)
class IOpResult(ISSAValue):
    """Represents an immutable SSA variable defined by an operation result."""

    op: IOperation
    index: int

    def __hash__(self) -> int:
        return hash(id(self.op)) + hash(self.index)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, IOpResult):
            return self.op == other.op and self.index == other.index
        return False


@dataclass(frozen=True)
class IBlockArg(ISSAValue):
    """Represents an immutable SSA variable defined by a basic block."""

    block: IBlock
    index: int

    def __hash__(self) -> int:
        return hash(id(self.block)) + hash(self.index)

    def __eq__(self, other: object) -> bool:
        return self is other

    def __repr__(self) -> str:
        return "BlockArg(type:" + self.type.name + ("attached") + ")"


@dataclass(frozen=True)
class IRegion:
    """An immutable region contains a CFG of immutable blocks. IRegions are contained in operations."""

    blocks: IList[IBlock]
    """Immutable blocks contained in the IRegion. The first block is the entry block."""

    def __hash__(self) -> int:
        return hash(id(self))

    def __eq__(self, value: object) -> bool:
        return self is value

    @property
    def block(self) -> IBlock:
        """
        Returns the block of a single-block region.
        Returns an exception if the region is not single-block.
        """
        if len(self.blocks) != 1:
            raise ValueError(
                "'block' property of IRegion class is only available "
                "for single-block regions."
            )
        return self.blocks[0]

    @property
    def ops(self) -> IList[IOperation]:
        """
        Get the operations of a single-block region.
        Returns an exception if the region is not single-block.
        """
        if len(self.blocks) != 1:
            raise ValueError(
                "'ops' property of IRegion class is only available "
                "for single-block regions."
            )
        return self.block.ops

    def __init__(self, blocks: Sequence[IBlock]):
        """Creates a new immutable region from a sequence of immutable blocks."""

        object.__setattr__(self, "blocks", IList(blocks))
        self.blocks.freeze()

    @classmethod
    def from_mutable(
        cls,
        blocks: Iterable[Block],
        value_map: dict[SSAValue, ISSAValue] | None = None,
        block_map: dict[Block, IBlock] | None = None,
    ) -> IRegion:
        """
        Creates a new immutable region from a sequence of mutable blocks.
        The value_map and block_map are used to map already known correspondings
        of mutable values to immutable values and mutable blocks to immutable blocks.
        """
        if value_map is None:
            value_map = {}
        if block_map is None:
            block_map = {}

        # adding dummy block mappings so that ops have a successor to reference
        # when the actual block is created all successor references will be moved
        # to the correct block
        for block in blocks:
            block_map[block] = IBlock([], [])

        immutable_blocks = [
            IBlock.from_mutable(block, value_map, block_map) for block in blocks
        ]

        region = IRegion(immutable_blocks)

        # clean up successor references to blocks for ops inside this region
        for block, imm_block in zip(blocks, region.blocks):
            if block.parent is None:
                raise InvalidIRException(
                    "Cannot create an IRegion from a mutable Block "
                    "that is not attached to a Region."
                )
            dummy_block = block_map[block]
            for block in region.blocks:
                for op in block.ops:
                    if dummy_block in op.successors:
                        dummy_index = op.successors.index(dummy_block)
                        # replace dummy successor with actual successor
                        object.__setattr__(
                            op,
                            "successors",
                            IList(
                                op.successors[:dummy_index]
                                + [imm_block]
                                + op.successors[dummy_index + 1 :]
                            ),
                        )

        return region

    def to_mutable(
        self,
        value_mapping: dict[ISSAValue, SSAValue] | None = None,
        block_mapping: dict[IBlock, Block] | None = None,
    ) -> Region:
        """
        Returns a mutable region that is a copy of this immutable region.
        The value_mapping and block_mapping are used to map already known correspondings
        of immutable values to mutable values and immutable blocks to mutable blocks.
        """
        if value_mapping is None:
            value_mapping = {}
        if block_mapping is None:
            block_mapping = {}
        mutable_blocks: list[Block] = []
        # All mutable blocks have to be initialized first so that ops can
        # refer to them in their successor lists.
        for block in self.blocks:
            mutable_blocks.append(mutable_block := Block(arg_types=block.arg_types))
            block_mapping[block] = mutable_block
        for block in self.blocks:
            # This will use the already created Block and populate it
            block.to_mutable(value_mapping=value_mapping, block_mapping=block_mapping)
        return Region(mutable_blocks)


@dataclass(frozen=True)
class IBlock:
    """An immutable block contains a list of immutable operations. IBlocks are contained in IRegions."""

    args: IList[IBlockArg]
    ops: IList[IOperation]

    @property
    def arg_types(self) -> list[Attribute]:
        frozen_arg_types = [arg.type for arg in self.args]
        return frozen_arg_types

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, value: object) -> bool:
        return self is value

    def __repr__(self) -> str:
        return (
            "block of" + str(len(self.ops)) + " operations with args: " + str(self.args)
        )

    def __post_init__(self):
        for arg in self.args:
            object.__setattr__(arg, "block", self)

    def __init__(
        self, args: Sequence[Attribute] | Sequence[IBlockArg], ops: Sequence[IOperation]
    ):
        """Creates a new immutable block."""

        # Type Guards:
        def is_iblock_arg_seq(list: Sequence[Any]) -> TypeGuard[Sequence[IBlockArg]]:
            return all([isinstance(elem, IBlockArg) for elem in list])

        def is_type_seq(list: Sequence[Any]) -> TypeGuard[Sequence[Attribute]]:
            return all([isinstance(elem, Attribute) for elem in list])

        block_args: Sequence[IBlockArg]
        if is_type_seq(args):
            block_args = [
                IBlockArg(type, IList(()), self, idx) for idx, type in enumerate(args)
            ]
        elif is_iblock_arg_seq(args):
            block_args = args
            for block_arg in block_args:
                object.__setattr__(block_arg, "block", self)
        else:
            raise Exception("args for IBlock ill structured")

        object.__setattr__(self, "args", IList(block_args))
        object.__setattr__(self, "ops", IList(ops))

        self.args.freeze()
        self.ops.freeze()

    @classmethod
    def from_mutable(
        cls,
        block: Block,
        value_map: dict[SSAValue, ISSAValue] | None = None,
        block_map: dict[Block, IBlock] | None = None,
    ) -> IBlock:
        """
        Creates an immutable block from a mutable block.
        The value_map and block_map are used to map already known correspondings
        of mutable values to immutable values and mutable blocks to immutable blocks.
        """
        if value_map is None:
            value_map = {}
        if block_map is None:
            block_map = {}

        args: list[IBlockArg] = []
        for arg in block.args:
            # The IBlock that will house this IBlockArg is not constructed yet.
            # After construction the block field will be set by the IBlock.
            immutable_arg = IBlockArg(
                arg.type,
                IList(),
                cast(IBlock, None),
                arg.index,
            )
            args.append(immutable_arg)
            value_map[arg] = immutable_arg

        immutable_ops = [
            IOperation.from_mutable(
                op, value_map=value_map, block_map=block_map, existing_operands=None
            )
            for op in block.ops
        ]

        return IBlock(args, immutable_ops)

    def to_mutable(
        self,
        value_mapping: dict[ISSAValue, SSAValue] | None = None,
        block_mapping: dict[IBlock, Block] | None = None,
    ) -> Block:
        """
        Returns a mutable block that is a copy of this immutable block.
        The value_mapping and block_mapping are used to map already known correspondings
        of immutable values to mutable values and immutable blocks to mutable blocks.
        """
        if value_mapping is None:
            value_mapping = {}
        if block_mapping is None:
            block_mapping = {}

        # Block might already have been created by the Region, look it up
        if self in block_mapping:
            mutable_block = block_mapping[self]
        else:
            mutable_block = Block(arg_types=self.arg_types)
        for idx, arg in enumerate(self.args):
            value_mapping[arg] = mutable_block.args[idx]
        block_mapping[self] = mutable_block

        for immutable_op in self.ops:
            mutable_block.add_op(
                immutable_op.to_mutable(
                    value_mapping=value_mapping, block_mapping=block_mapping
                )
            )
        return mutable_block


def get_immutable_copy(op: Operation) -> IOperation:
    return IOperation.from_mutable(op)


@dataclass(frozen=True)
class IOperation:
    """Represents an immutable operation."""

    __match_args__ = ("op_type", "operands", "results", "successors", "regions")
    name: str
    op_type: type[Operation]
    properties: immutabledict[str, Attribute]
    attributes: immutabledict[str, Attribute]
    operands: IList[ISSAValue]
    results: IList[IOpResult]
    successors: IList[IBlock]
    regions: IList[IRegion]

    def __init__(
        self,
        name: str,
        op_type: type[Operation],
        attributes: immutabledict[str, Attribute],
        properties: immutabledict[str, Attribute],
        operands: Sequence[ISSAValue],
        result_types: Sequence[Attribute],
        successors: Sequence[IBlock],
        regions: Sequence[IRegion],
    ) -> None:
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "op_type", op_type)
        object.__setattr__(self, "attributes", attributes)
        object.__setattr__(self, "properties", properties)
        object.__setattr__(self, "operands", IList(operands))
        for operand in operands:
            operand._add_user(self)  # pyright: ignore[reportPrivateUsage]
        object.__setattr__(
            self,
            "results",
            IList(
                [
                    IOpResult(type, IList(()), self, idx)
                    for idx, type in enumerate(result_types)
                ]
            ),
        )
        object.__setattr__(self, "successors", IList(successors))
        object.__setattr__(self, "regions", IList(regions))

        self.operands.freeze()
        self.results.freeze()
        self.successors.freeze()
        self.regions.freeze()

    @classmethod
    def get(
        cls,
        name: str,
        op_type: type[Operation],
        operands: Sequence[ISSAValue],
        result_types: Sequence[Attribute],
        attributes: immutabledict[str, Attribute],
        properties: immutabledict[str, Attribute],
        successors: Sequence[IBlock],
        regions: Sequence[IRegion],
    ) -> IOperation:
        return cls(
            name,
            op_type,
            properties,
            attributes,
            operands,
            result_types,
            successors,
            regions,
        )

    def __hash__(self) -> int:
        return hash(id(self))

    def __eq__(self, value: object, /) -> bool:
        return self is value

    @property
    def result(self) -> IOpResult:
        """
        Get the result of a of an IOperation with a single result.
        Returns an exception if the operation does not have exactly one result.
        """
        if len(self.results) != 1:
            raise ValueError(
                "'result' property of IOperation class is only available "
                "for IOperations with exactly one result."
            )
        return self.results[0]

    @property
    def region(self) -> IRegion:
        """
        Get the region of a of an IOperation with a single region.
        Returns an exception if the operation does not have exactly one region.
        """
        if len(self.regions) != 1:
            raise ValueError(
                "'region' property of IOperation class is only available "
                "for IOperations with exactly one region."
            )
        return self.regions[0]

    @property
    def result_types(self) -> list[Attribute]:
        return [result.type for result in self.results]

    def to_mutable(
        self,
        value_mapping: dict[ISSAValue, SSAValue] | None = None,
        block_mapping: dict[IBlock, Block] | None = None,
    ) -> Operation:
        """
        Returns a mutable operation that is a copy of this immutable operation.
        The value_mapping and block_mapping are used to map already known correspondings
        of immutable values to mutable values and immutable blocks to mutable blocks.
        """
        if value_mapping is None:
            value_mapping = {}
        if block_mapping is None:
            block_mapping = {}

        mutable_operands: list[SSAValue] = []
        for operand in self.operands:
            if operand in value_mapping:
                mutable_operands.append(value_mapping[operand])
            else:
                print(f"ERROR: op {self.name} uses SSAValue before definition")
                # Continuing to enable printing the IR including missing
                # operands for investigation
                mutable_operands.append(
                    OpResult(operand.type, cast(Operation, None), 0)
                )

        mutable_successors: list[Block] = []
        for successor in self.successors:
            if successor in block_mapping:
                mutable_successors.append(block_mapping[successor])
            else:
                raise InvalidIRException(
                    "Invalid IR: Block is not defined in the current region"
                )

        mutable_regions: list[Region] = []
        for region in self.regions:
            mutable_regions.append(
                region.to_mutable(
                    value_mapping=value_mapping, block_mapping=block_mapping
                )
            )

        new_op: Operation = self.op_type.create(
            operands=mutable_operands,
            result_types=[result.type for result in self.results],
            properties=dict(self.properties),
            attributes=dict(self.attributes),
            successors=mutable_successors,
            regions=mutable_regions,
        )

        # Add the results of this operation to the value mapping
        # so other operations can use them as operands.
        for idx, result in enumerate(self.results):
            m_result = new_op.results[idx]
            value_mapping[result] = m_result

        return new_op

    @classmethod
    def from_mutable(
        cls,
        op: Operation,
        value_map: dict[SSAValue, ISSAValue] | None = None,
        block_map: dict[Block, IBlock] | None = None,
        existing_operands: Sequence[ISSAValue] | None = None,
    ) -> IOperation:
        """
        Returns an immutable operation that is a copy of the given mutable operation.
        The value_map and block_map are used to map already known correspondings
        of mutable values to immutable values and mutable blocks to immutable blocks.
        """
        assert isinstance(op, Operation)
        op_type = op.__class__

        if value_map is None:
            value_map = {}
        if block_map is None:
            block_map = {}

        operands: list[ISSAValue] = []
        if existing_operands is None:
            for operand in op.operands:
                if isinstance(operand, OpResult):
                    if operand in value_map:
                        operands.append(value_map[operand])
                    else:
                        raise Exception("Operand used before definition")
                elif isinstance(operand, BlockArgument):
                    if operand not in value_map:
                        raise Exception(
                            "Block argument expected in mapping for op: " + op.name
                        )
                    operands.append(value_map[operand])
                else:
                    raise Exception(
                        "Operand is expected to be either OpResult or BlockArgument"
                    )
        else:
            operands.extend(existing_operands)

        properties: immutabledict[str, Attribute] = immutabledict(op.properties)
        attributes: immutabledict[str, Attribute] = immutabledict(op.attributes)

        successors: list[IBlock] = []
        for successor in op.successors:
            if successor in block_map:
                successors.append(block_map[successor])
            else:
                raise Exception(
                    "Successor not defined in current region, `from_mutable`\
                          probably has to be called on the parent operation."
                )

        regions: list[IRegion] = []
        for region in op.regions:
            regions.append(IRegion.from_mutable(region.blocks, value_map, block_map))

        immutable_op = IOperation.get(
            op.name,
            op_type,
            operands,
            op.result_types,
            properties,
            attributes,
            successors,
            regions,
        )

        for idx, result in enumerate(op.results):
            value_map[result] = immutable_op.results[idx]

        return immutable_op

    def get_attribute(self, name: str) -> Attribute:
        return self.attributes[name]

    def get_attributes_copy(self) -> dict[str, Attribute]:
        return dict(self.attributes)



from typing import cast

from xdsl.context import Context
from xdsl.dialects import arith, builtin, scf
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.transforms.canonicalization_patterns.utils import const_evaluate_operand

#  This pass flattens pairs nested loops into a single loop.
#
#  Similar to LLVM's loop flattening:
#  https://llvm.org/doxygen/LoopFlatten_8cpp_source.html
#
#  The intention is to optimise loop nests like this, which together access an
#  array linearly:
#
#    for i in range(x, N, M):
#      for j in range(0, M, K):
#        f(A[i+j])
#
#    for o in range(ol, ou, os):
#      for i in range(il, iu, is):
#        # neither o nor i are used
#
#    These become:
#    # (If K is constant and divides M)
#    for i in range(x, N, K):
#      f(A[i])
#
#    factor = (iu - il) // is
#    for o in range(ol, ou * factor, os):
#      # o is not used
#


class FlattenNestedLoopsPattern(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: scf.ForOp, rewriter: PatternRewriter) -> None:
        outer_body = op.body.block
        if not isinstance(inner_loop := outer_body.first_op, scf.ForOp):
            # Outer loop must contain inner loop
            return
        if (
            inner_loop
            is not (outer_yield_op := cast(scf.YieldOp, outer_body.last_op)).prev_op
        ):
            # Outer loop must contain only inner loop and yield
            return

        if op.iter_args:
            if not inner_loop.iter_args:
                return
            if len(op.iter_args) != len(inner_loop.iter_args):
                return
            if not all(
                lhs is rhs
                for (lhs, rhs) in zip(outer_body.args[1:], inner_loop.iter_args)
            ):
                return
            if not all(
                lhs is rhs
                for (lhs, rhs) in zip(inner_loop.results, outer_yield_op.operands)
            ):
                return
        elif inner_loop.iter_args:
            return

        if (inner_lb := const_evaluate_operand(inner_loop.lb)) is None:
            return

        if (inner_ub := const_evaluate_operand(inner_loop.ub)) is None:
            return
        if (outer_step := const_evaluate_operand(op.step)) is None:
            return
        if (inner_step := const_evaluate_operand(inner_loop.step)) is None:
            return

        outer_index = outer_body.args[0]
        inner_index = inner_loop.body.block.args[0]

        if outer_index.uses or inner_index.uses:
            if inner_lb != 0:
                return

            if inner_ub != outer_step:
                return

            if outer_step % inner_step:
                return

            # If either induction variable is used, we can only fold if used exactly once
            if not outer_index.has_one_use() or not inner_index.has_one_use():
                return

            outer_user = next(iter(outer_index.uses)).operation
            inner_user = next(iter(inner_index.uses)).operation
            if outer_user is not inner_user:
                return

            user = outer_user

            if not isinstance(user, arith.AddiOp):
                return

            # We can fuse
            user.result.replace_by(inner_index)
            rewriter.erase_op(user)
            new_ub = op.ub
            new_step = inner_loop.step
        else:
            if (outer_lb := const_evaluate_operand(op.lb)) is None:
                return

            if outer_lb != 0:
                # Do not currently handle lb != 0
                return

            factor = (inner_ub - inner_lb) // inner_step
            factor_op = arith.ConstantOp(
                builtin.IntegerAttr(factor, builtin.IndexType())
            )
            new_ub_op = arith.MuliOp(op.ub, factor_op.result)
            rewriter.insert_op_before_matched_op((factor_op, new_ub_op))
            new_ub = new_ub_op.result
            new_step = op.step

        moved_region = rewriter.move_region_contents_to_new_regions(inner_loop.body)
        rewriter.erase_op(outer_yield_op)
        rewriter.erase_op(inner_loop)

        rewriter.replace_matched_op(
            scf.ForOp(
                op.lb,
                new_ub,
                new_step,
                op.iter_args,
                moved_region,
            )
        )


class ScfForLoopFlattenPass(ModulePass):
    """
    Folds perfect loop nests if they can be represented with a single loop.
    Currently does this by matching the inner loop range with the outer loop step.
    If the inner iteration space fits perfectly in the outer iteration step, then merge.
    Other conditions:
     - the only use of the induction arguments must be an add operation, this op is fused
       into a single induction argument,
     - the lower bound of the inner loop must be 0,
     - the loops must have no iteration arguments.
    """

    name = "scf-for-loop-flatten"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(FlattenNestedLoopsPattern()).rewrite_module(op)



from xdsl.dialects import builtin, stencil
from xdsl.dialects.experimental import dmp
from xdsl.passes import Context, ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


class CanonicalizeDmpSwap(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: dmp.SwapOp, rewriter: PatternRewriter, /):
        keeps: list[dmp.ExchangeDeclarationAttr] = []
        for swap in op.swaps:
            if swap.elem_count > 0:
                keeps.append(swap)
        if len(keeps) == 0:
            new_result = (
                op.input_stencil
                if isinstance(op.input_stencil.type, stencil.TempType)
                else None
            )
            rewriter.replace_matched_op([], [new_result])
        else:
            op.swaps = builtin.ArrayAttr(keeps)


class CanonicalizeDmpPass(ModulePass):
    name = "canonicalize-dmp"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(CanonicalizeDmpSwap()).rewrite_module(op)



from collections.abc import Callable, Sequence
from typing import cast

from xdsl.dialects import arith, cf
from xdsl.dialects.builtin import (
    BoolAttr,
    DenseIntElementsAttr,
    IntegerAttr,
    VectorType,
)
from xdsl.ir import Block, BlockArgument, Operation, SSAValue
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.transforms.canonicalization_patterns.utils import (
    const_evaluate_operand,
)


class AssertTrue(RewritePattern):
    """Erase assertion if argument is constant true."""

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.AssertOp, rewriter: PatternRewriter):
        owner = op.arg.owner

        if not isinstance(owner, arith.ConstantOp):
            return

        value = owner.value

        if not isinstance(value, IntegerAttr):
            return

        if not value.value.data:
            return

        rewriter.replace_matched_op([])


class SimplifyBrToBlockWithSinglePred(RewritePattern):
    """
    Simplify a branch to a block that has a single predecessor. This effectively
    merges the two blocks.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.BranchOp, rewriter: PatternRewriter):
        succ = op.successor
        parent = op.parent_block()
        if parent is None:
            return

        # Check that the successor block has a single predecessor
        if succ == parent or len(succ.predecessors()) != 1:
            return

        br_operands = op.operands
        rewriter.erase_matched_op()
        rewriter.inline_block(succ, InsertPoint.at_end(parent), br_operands)


def collapse_branch(
    successor: Block, successor_operands: Sequence[SSAValue]
) -> tuple[Block, Sequence[SSAValue]] | None:
    """
    Given a successor, try to collapse it to a new destination if it only
    contains a passthrough unconditional branch. If the successor is
    collapsable, `successor` and `successorOperands` are updated to reference
    the new destination and values. `argStorage` is used as storage if operands
    to the collapsed successor need to be remapped. It must outlive uses of
    successorOperands.
    """

    # Check that successor only contains branch
    if len(successor.ops) != 1:
        return

    branch = successor.ops.first
    # Check that the terminator is an unconditional branch
    if not isinstance(branch, cf.BranchOp):
        return

    # Check that the arguments are only used within the terminator
    for argument in successor.args:
        for user in argument.uses:
            if user.operation != branch:
                return

    # Don't try to collapse branches to infinite loops.
    if branch.successor == successor:
        return

    # Remap operands
    operands = branch.operands

    new_operands = tuple(
        successor_operands[operand.index]
        if isinstance(operand, BlockArgument) and operand.owner is successor
        else operand
        for operand in operands
    )

    return (branch.successor, new_operands)


class SimplifyPassThroughBr(RewritePattern):
    """
      br ^bb1
    ^bb1
      br ^bbN(...)

     -> br ^bbN(...)
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.BranchOp, rewriter: PatternRewriter):
        # Check the successor doesn't point back to the current block
        parent = op.parent_block()
        if parent is None or op.successor == parent:
            return

        ret = collapse_branch(op.successor, op.arguments)
        if ret is None:
            return
        (block, args) = ret

        rewriter.replace_matched_op(cf.BranchOp(block, *args))


class SimplifyConstCondBranchPred(RewritePattern):
    """
    cf.cond_br true, ^bb1, ^bb2
     -> br ^bb1
    cf.cond_br false, ^bb1, ^bb2
     -> br ^bb2
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.ConditionalBranchOp, rewriter: PatternRewriter):
        # Check if cond operand is constant
        cond = const_evaluate_operand(op.cond)

        if cond is None:
            return

        if cond:
            rewriter.replace_matched_op(cf.BranchOp(op.then_block, *op.then_arguments))
        else:
            rewriter.replace_matched_op(cf.BranchOp(op.else_block, *op.else_arguments))


class SimplifyPassThroughCondBranch(RewritePattern):
    """
      cf.cond_br %cond, ^bb1, ^bb2
    ^bb1
      br ^bbN(...)
    ^bb2
      br ^bbK(...)

     -> cf.cond_br %cond, ^bbN(...), ^bbK(...)
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.ConditionalBranchOp, rewriter: PatternRewriter):
        # Try to collapse both branches
        collapsed_then = collapse_branch(op.then_block, op.then_arguments)
        collapsed_else = collapse_branch(op.else_block, op.else_arguments)

        # If neither collapsed then we return
        if collapsed_then is None and collapsed_else is None:
            return

        (new_then, new_then_args) = collapsed_then or (op.then_block, op.then_arguments)

        (new_else, new_else_args) = collapsed_else or (op.else_block, op.else_arguments)

        rewriter.replace_matched_op(
            cf.ConditionalBranchOp(
                op.cond, new_then, new_then_args, new_else, new_else_args
            )
        )


class SimplifyCondBranchIdenticalSuccessors(RewritePattern):
    """
    cf.cond_br %cond, ^bb1(A, ..., N), ^bb1(A, ..., N)
     -> br ^bb1(A, ..., N)

    cf.cond_br %cond, ^bb1(A), ^bb1(B)
     -> %select = arith.select %cond, A, B
        br ^bb1(%select)
    """

    @staticmethod
    def _merge_operand(
        op1: SSAValue,
        op2: SSAValue,
        rewriter: PatternRewriter,
        cond_br: cf.ConditionalBranchOp,
    ) -> SSAValue:
        if op1 == op2:
            return op1
        select = arith.SelectOp(cond_br.cond, op1, op2)
        rewriter.insert_op(select, InsertPoint.before(cond_br))
        return select.result

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.ConditionalBranchOp, rewriter: PatternRewriter):
        # Check that the true and false destinations are the same
        if op.then_block != op.else_block:
            return

        merged_operands = tuple(
            self._merge_operand(op1, op2, rewriter, op)
            for (op1, op2) in zip(op.then_arguments, op.else_arguments, strict=True)
        )

        rewriter.replace_matched_op(cf.BranchOp(op.then_block, *merged_operands))


class CondBranchTruthPropagation(RewritePattern):
    """
      cf.cond_br %arg0, ^trueB, ^falseB

    ^trueB:
      "test.consumer1"(%arg0) : (i1) -> ()
       ...

    ^falseB:
      "test.consumer2"(%arg0) : (i1) -> ()
      ...

    ->

      cf.cond_br %arg0, ^trueB, ^falseB
    ^trueB:
      "test.consumer1"(%true) : (i1) -> ()
      ...

    ^falseB:
      "test.consumer2"(%false) : (i1) -> ()
      ...
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.ConditionalBranchOp, rewriter: PatternRewriter):
        if len(op.then_block.predecessors()) == 1:
            if any(
                use.operation.parent_block() is op.then_block for use in op.cond.uses
            ):
                const_true = arith.ConstantOp(BoolAttr.from_bool(True))
                rewriter.insert_op(const_true, InsertPoint.before(op))
                op.cond.replace_by_if(
                    const_true.result,
                    lambda use: use.operation.parent_block() is op.then_block,
                )
        if len(op.else_block.predecessors()) == 1:
            if any(
                use.operation.parent_block() is op.else_block for use in op.cond.uses
            ):
                const_false = arith.ConstantOp(BoolAttr.from_bool(False))
                rewriter.insert_op(const_false, InsertPoint.before(op))
                op.cond.replace_by_if(
                    const_false.result,
                    lambda use: use.operation.parent_block() is op.else_block,
                )


class SimplifySwitchWithOnlyDefault(RewritePattern):
    """
    switch %flag : i32, [
      default:  ^bb1
    ]
     -> br ^bb1
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.SwitchOp, rewriter: PatternRewriter):
        if not op.case_blocks:
            rewriter.replace_matched_op(
                cf.BranchOp(op.default_block, *op.default_operands)
            )


def drop_case_helper(
    rewriter: PatternRewriter,
    op: cf.SwitchOp,
    predicate: Callable[[IntegerAttr, Block, Sequence[Operation | SSAValue]], bool],
):
    case_values = op.case_values
    if case_values is None:
        return
    requires_change = False

    new_case_values: list[int] = []
    new_case_blocks: list[Block] = []
    new_case_operands: list[Sequence[Operation | SSAValue]] = []

    for switch_case, block, operands in zip(
        case_values.get_attrs(),
        op.case_blocks,
        op.case_operand,
        strict=True,
    ):
        int_switch_case = cast(IntegerAttr, switch_case)
        if predicate(int_switch_case, block, operands):
            requires_change = True
            continue
        new_case_values.append(cast(IntegerAttr, switch_case).value.data)
        new_case_blocks.append(block)
        new_case_operands.append(operands)

    if requires_change:
        rewriter.replace_matched_op(
            cf.SwitchOp(
                op.flag,
                op.default_block,
                op.default_operands,
                DenseIntElementsAttr.from_list(
                    VectorType(case_values.get_element_type(), (len(new_case_values),)),
                    new_case_values,
                ),
                new_case_blocks,
                new_case_operands,
            )
        )


class DropSwitchCasesThatMatchDefault(RewritePattern):
    """
    switch %flag : i32, [
      default: ^bb1,
      42: ^bb1,
      43: ^bb2
    ]
    ->
    switch %flag : i32, [
      default: ^bb1,
      43: ^bb2
    ]
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.SwitchOp, rewriter: PatternRewriter):
        def predicate(
            switch_case: IntegerAttr,
            block: Block,
            operands: Sequence[Operation | SSAValue],
        ) -> bool:
            return block == op.default_block and operands == op.default_operands

        drop_case_helper(rewriter, op, predicate)


def fold_switch(switch: cf.SwitchOp, rewriter: PatternRewriter, flag: int):
    """
    Helper for folding a switch with a constant value.
    switch %c_42 : i32, [
      default: ^bb1 ,
      42: ^bb2,
      43: ^bb3
    ]
    -> br ^bb2
    """
    case_values = () if switch.case_values is None else switch.case_values.get_attrs()

    new_block, new_operands = next(
        (
            (block, operand)
            for (c, block, operand) in zip(
                case_values, switch.case_blocks, switch.case_operand, strict=True
            )
            if flag == c.value.data
        ),
        (switch.default_block, switch.default_operands),
    )

    rewriter.replace_matched_op(cf.BranchOp(new_block, *new_operands))


class SimplifyConstSwitchValue(RewritePattern):
    """
    switch %c_42 : i32, [
      default: ^bb1,
      42: ^bb2,
      43: ^bb3
    ]
    -> br ^bb2
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.SwitchOp, rewriter: PatternRewriter):
        if (flag := const_evaluate_operand(op.flag)) is not None:
            fold_switch(op, rewriter, flag)


class SimplifyPassThroughSwitch(RewritePattern):
    """
    switch %c_42 : i32, [
      default: ^bb1,
      42: ^bb2,
    ]
    ^bb2:
      br ^bb3
    ->
    switch %c_42 : i32, [
      default: ^bb1,
      42: ^bb3,
    ]
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.SwitchOp, rewriter: PatternRewriter):
        requires_change = False

        new_case_blocks: list[Block] = []
        new_case_operands: list[Sequence[Operation | SSAValue]] = []

        for block, operands in zip(op.case_blocks, op.case_operand, strict=True):
            collapsed = collapse_branch(block, operands)
            requires_change |= collapsed is not None
            (new_block, new_operands) = collapsed or (block, operands)
            new_case_blocks.append(new_block)
            new_case_operands.append(new_operands)

        collapsed = collapse_branch(op.default_block, op.default_operands)

        requires_change |= collapsed is not None

        (default_block, default_operands) = collapsed or (
            op.default_block,
            op.default_operands,
        )

        if requires_change:
            rewriter.replace_matched_op(
                cf.SwitchOp(
                    op.flag,
                    default_block,
                    default_operands,
                    op.case_values,
                    new_case_blocks,
                    new_case_operands,
                )
            )


class SimplifySwitchFromSwitchOnSameCondition(RewritePattern):
    """
    switch %flag : i32, [
      default: ^bb1,
      42: ^bb2,
    ]
    ^bb2:
      switch %flag : i32, [
        default: ^bb3,
        42: ^bb4
      ]
    ->
    switch %flag : i32, [
      default: ^bb1,
      42: ^bb2,
    ]
    ^bb2:
      br ^bb4

     and

    switch %flag : i32, [
      default: ^bb1,
      42: ^bb2,
    ]
    ^bb2:
      switch %flag : i32, [
        default: ^bb3,
        43: ^bb4
      ]
    ->
    switch %flag : i32, [
      default: ^bb1,
      42: ^bb2,
    ]
    ^bb2:
      br ^bb3

    and

    switch %flag : i32, [
      default: ^bb1,
      42: ^bb2
    ]
    ^bb1:
      switch %flag : i32, [
        default: ^bb3,
        42: ^bb4,
        43: ^bb5
      ]
    ->
    switch %flag : i32, [
      default: ^bb1,
      42: ^bb2,
    ]
    ^bb1:
      switch %flag : i32, [
        default: ^bb3,
        43: ^bb5
      ]
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: cf.SwitchOp, rewriter: PatternRewriter):
        block = op.parent_block()
        if block is None:
            return
        if (pred := block.get_unique_use()) is None:
            return
        switch = pred.operation
        if not isinstance(switch, cf.SwitchOp):
            return

        if switch.flag != op.flag:
            return

        case_values = switch.case_values
        if case_values is None:
            return

        if pred.index != 0:
            fold_switch(
                op,
                rewriter,
                case_values.get_values()[pred.index - 1],
            )
        else:

            def predicate(
                switch_case: IntegerAttr,
                block: Block,
                operands: Sequence[Operation | SSAValue],
            ) -> bool:
                return switch_case in case_values.get_attrs()

            drop_case_helper(rewriter, op, predicate)
